from __future__ import annotations

import asyncio
import functools
import json
import logging
import math
import stat
import uuid
from datetime import datetime
from pathlib import Path
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Tuple,
    Type,
)

import aiohttp
import aiohttp_cors
import aiotools
import attrs
import sqlalchemy as sa
import trafaret as t
from aiohttp import web

from ai.backend.common import validators as tx
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import VFolderHostPermission, VFolderHostPermissionMap

from ..models import (
    AgentStatus,
    KernelStatus,
    UserRole,
    VFolderAccessStatus,
    VFolderCloneInfo,
    VFolderDeletionInfo,
    VFolderInvitationState,
    VFolderOperationStatus,
    VFolderOwnershipType,
    VFolderPermission,
    VFolderPermissionValidator,
    VFolderUsageMode,
    agents,
    ensure_host_permission_allowed,
    filter_host_allowed_permission,
    get_allowed_vfolder_hosts_by_group,
    get_allowed_vfolder_hosts_by_user,
    groups,
    initiate_vfolder_clone,
    initiate_vfolder_removal,
    kernels,
    keypairs,
    query_accessible_vfolders,
    query_owned_dotfiles,
    users,
    verify_vfolder_name,
    vfolder_invitations,
    vfolder_permissions,
    vfolders,
)
from .auth import admin_required, auth_required, superadmin_required
from .exceptions import (
    BackendAgentError,
    GenericForbidden,
    GroupNotFound,
    InsufficientPrivilege,
    InternalServerError,
    InvalidAPIParameters,
    ObjectNotFound,
    ServerMisconfiguredError,
    TooManyVFoldersFound,
    VFolderAlreadyExists,
    VFolderCreationFailed,
    VFolderFilterStatusFailed,
    VFolderFilterStatusNotAvailable,
    VFolderNotFound,
)
from .manager import ALL_ALLOWED, READ_ALLOWED, server_status_required
from .resource import get_watcher_info
from .utils import check_api_params, get_user_scopes

if TYPE_CHECKING:
    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]

VFolderRow = Mapping[str, Any]


async def ensure_vfolder_status(
    request: web.Request,
    perm: VFolderAccessStatus,
    folder_id: Optional[str] = None,
    folder_name: Optional[str] = None,
):
    """
    Checks if the target vfolder status is READY.
    This function should prevent user access
    while storage-proxy operations such as vfolder clone, deletion is handling.
    """

    root_ctx: RootContext = request.app["_root.context"]
    domain_name = request["user"]["domain_name"]
    user_role = request["user"]["role"]
    user_uuid = request["user"]["uuid"]

    allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
    if folder_id:
        vf_name_conds = vfolders.c.id == folder_id
    elif folder_name:
        vf_name_conds = vfolders.c.name == folder_name
    else:
        raise VFolderFilterStatusFailed("either vFolder id nor name not supplied")

    available_vf_statuses = set()
    match perm:
        case VFolderAccessStatus.READABLE:
            # if READABLE access status is requested, all operation statuses are accepted.
            available_vf_statuses = {
                VFolderOperationStatus.READY,
                VFolderOperationStatus.PERFORMING,
                VFolderOperationStatus.CLONING,
                VFolderOperationStatus.DELETING,
                VFolderOperationStatus.MOUNTED,
            }
        case VFolderAccessStatus.UPDATABLE:
            # if UPDATABLE access status is requested, READY and MOUNTED operation statuses are accepted.
            available_vf_statuses = {
                VFolderOperationStatus.READY,
                VFolderOperationStatus.MOUNTED,
            }
        case VFolderAccessStatus.DELETABLE:
            # if DELETABLE access status is requested, only READY operation status is accepted.
            available_vf_statuses = {
                VFolderOperationStatus.READY,
            }
        case _:
            # Otherwise, raise VFolderFilterStatusNotAvailable()
            raise VFolderFilterStatusNotAvailable()
    async with root_ctx.db.begin_readonly() as conn:
        entries = await query_accessible_vfolders(
            conn,
            user_uuid,
            user_role=user_role,
            domain_name=domain_name,
            allowed_vfolder_types=allowed_vfolder_types,
            extra_vf_conds=vf_name_conds,
            allow_privileged_access=True,
        )
        if len(entries) == 0:
            raise VFolderFilterStatusFailed(
                f"Cannot find any folder with the given identity ({folder_id = }, {folder_name = })"
            )
        for entry in entries:
            if entry["status"] not in available_vf_statuses:
                raise VFolderFilterStatusFailed()


def vfolder_permission_required(perm: VFolderPermission):
    """
    Checks if the target vfolder exists and is either:
    - owned by the current access key, or
    - allowed accesses by the access key under the specified permission.

    The decorated handler should accept an extra argument
    which contains a dict object describing the matched VirtualFolder table row.
    """

    # FIXME: replace ... with [web.Request, VFolderRow, Any...] in the future mypy
    def _wrapper(handler: Callable[..., Awaitable[web.Response]]):
        @functools.wraps(handler)
        async def _wrapped(request: web.Request, *args, **kwargs) -> web.Response:
            root_ctx: RootContext = request.app["_root.context"]
            domain_name = request["user"]["domain_name"]
            user_role = request["user"]["role"]
            user_uuid = request["user"]["uuid"]
            folder_name = request.match_info["name"]
            allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
            vf_user_cond = None
            vf_group_cond = None
            if perm == VFolderPermission.READ_ONLY:
                # if READ_ONLY is requested, any permission accepts.
                invited_perm_cond = vfolder_permissions.c.permission.in_(
                    [
                        VFolderPermission.READ_ONLY,
                        VFolderPermission.READ_WRITE,
                        VFolderPermission.RW_DELETE,
                    ]
                )
                if not request["is_admin"]:
                    vf_group_cond = vfolders.c.permission.in_(
                        [
                            VFolderPermission.READ_ONLY,
                            VFolderPermission.READ_WRITE,
                            VFolderPermission.RW_DELETE,
                        ]
                    )
            elif perm == VFolderPermission.READ_WRITE:
                invited_perm_cond = vfolder_permissions.c.permission.in_(
                    [
                        VFolderPermission.READ_WRITE,
                        VFolderPermission.RW_DELETE,
                    ]
                )
                if not request["is_admin"]:
                    vf_group_cond = vfolders.c.permission.in_(
                        [
                            VFolderPermission.READ_WRITE,
                            VFolderPermission.RW_DELETE,
                        ]
                    )
            elif perm == VFolderPermission.RW_DELETE:
                # If RW_DELETE is requested, only RW_DELETE accepts.
                invited_perm_cond = vfolder_permissions.c.permission == VFolderPermission.RW_DELETE
                if not request["is_admin"]:
                    vf_group_cond = vfolders.c.permission == VFolderPermission.RW_DELETE
            else:
                # Otherwise, just compare it as-is (for future compatibility).
                invited_perm_cond = vfolder_permissions.c.permission == perm
                if not request["is_admin"]:
                    vf_group_cond = vfolders.c.permission == perm
            async with root_ctx.db.begin() as conn:
                entries = await query_accessible_vfolders(
                    conn,
                    user_uuid,
                    user_role=user_role,
                    domain_name=domain_name,
                    allowed_vfolder_types=allowed_vfolder_types,
                    extra_vf_conds=(vfolders.c.name == folder_name),
                    extra_invited_vf_conds=invited_perm_cond,
                    extra_vf_user_conds=vf_user_cond,
                    extra_vf_group_conds=vf_group_cond,
                )
                if len(entries) == 0:
                    raise VFolderNotFound("Your operation may be permission denied.")
            return await handler(request, entries[0], *args, **kwargs)

        return _wrapped

    return _wrapper


# FIXME: replace ... with [web.Request, VFolderRow, Any...] in the future mypy
def vfolder_check_exists(handler: Callable[..., Awaitable[web.Response]]):
    """
    Checks if the target vfolder exists and is owned by the current user.

    The decorated handler should accept an extra "row" argument
    which contains the matched VirtualFolder table row.
    """

    @functools.wraps(handler)
    async def _wrapped(request: web.Request, *args, **kwargs) -> web.Response:
        root_ctx: RootContext = request.app["_root.context"]
        user_uuid = request["user"]["uuid"]
        folder_name = request.match_info["name"]
        async with root_ctx.db.begin() as conn:
            j = sa.join(
                vfolders,
                vfolder_permissions,
                vfolders.c.id == vfolder_permissions.c.vfolder,
                isouter=True,
            )
            query = (
                sa.select("*")
                .select_from(j)
                .where(
                    ((vfolders.c.user == user_uuid) | (vfolder_permissions.c.user == user_uuid))
                    & (vfolders.c.name == folder_name)
                )
            )
            try:
                result = await conn.execute(query)
            except sa.exc.DataError:
                raise InvalidAPIParameters
            row = result.first()
            if row is None:
                raise VFolderNotFound()
        return await handler(request, row, *args, **kwargs)

    return _wrapped


@auth_required
@server_status_required(ALL_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("name"): tx.Slug(allow_dot=True),
            t.Key("host", default=None) >> "folder_host": t.String | t.Null,
            t.Key("usage_mode", default="general"): tx.Enum(VFolderUsageMode) | t.Null,
            t.Key("permission", default="rw"): tx.Enum(VFolderPermission) | t.Null,
            tx.AliasedKey(["unmanaged_path", "unmanagedPath"], default=None): t.String | t.Null,
            tx.AliasedKey(["group", "groupId", "group_id"], default=None): tx.UUID
            | t.String
            | t.Null,
            t.Key("quota", default=None): tx.BinarySize | t.Null,
            t.Key("cloneable", default=False): t.Bool,
        }
    ),
)
async def create(request: web.Request, params: Any) -> web.Response:
    resp: Dict[str, Any] = {}
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    user_role = request["user"]["role"]
    user_uuid = request["user"]["uuid"]
    resource_policy = request["keypair"]["resource_policy"]
    domain_name = request["user"]["domain_name"]
    group_id_or_name = params["group"]
    log.info(
        "VFOLDER.CREATE (email:{}, ak:{}, vf:{}, vfh:{}, umod:{}, perm:{})",
        request["user"]["email"],
        access_key,
        params["name"],
        params["folder_host"],
        params["usage_mode"].value,
        params["permission"].value,
    )
    folder_host = params["folder_host"]
    unmanaged_path = params["unmanaged_path"]
    # Check if user is trying to created unmanaged vFolder
    if unmanaged_path:
        # Approve only if user is Admin or Superadmin
        if user_role not in (UserRole.ADMIN, UserRole.SUPERADMIN):
            raise GenericForbidden("Insufficient permission")
    else:
        # Resolve host for the new virtual folder.
        if not folder_host:
            folder_host = await root_ctx.shared_config.etcd.get("volumes/default_host")
            if not folder_host:
                raise InvalidAPIParameters(
                    "You must specify the vfolder host "
                    "because the default host is not configured."
                )
    allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
    for vf_type in allowed_vfolder_types:
        if vf_type not in ("user", "group"):
            raise ServerMisconfiguredError(
                f"Invalid vfolder type(s): {str(allowed_vfolder_types)}."
                ' Only "user" or "group" is allowed.'
            )

    if not verify_vfolder_name(params["name"]):
        raise InvalidAPIParameters(f'{params["name"]} is reserved for internal operations.')
    if params["name"].startswith(".") and params["name"] != ".local":
        if params["group"] is not None:
            raise InvalidAPIParameters("dot-prefixed vfolders cannot be a group folder.")

    async with root_ctx.db.begin() as conn:
        # Convert group name to uuid if group name is given.
        if isinstance(group_id_or_name, str):
            query = (
                sa.select([groups.c.id])
                .select_from(groups)
                .where(groups.c.domain_name == domain_name)
                .where(groups.c.name == group_id_or_name)
            )
            group_id = await conn.scalar(query)
        else:
            group_id = group_id_or_name
        if not unmanaged_path:
            await ensure_host_permission_allowed(
                conn,
                folder_host,
                allowed_vfolder_types=allowed_vfolder_types,
                user_uuid=user_uuid,
                resource_policy=resource_policy,
                domain_name=domain_name,
                group_id=group_id,
                permission=VFolderHostPermission.CREATE,
            )

        # Check resource policy's max_vfolder_count
        if resource_policy["max_vfolder_count"] > 0:
            query = sa.select([sa.func.count()]).where(vfolders.c.user == user_uuid)
            result = await conn.scalar(query)
            if result >= resource_policy["max_vfolder_count"]:
                raise InvalidAPIParameters("You cannot create more vfolders.")

        # Limit vfolder size quota if it is larger than max_vfolder_size of the resource policy.
        max_vfolder_size = resource_policy.get("max_vfolder_size", 0)
        if max_vfolder_size > 0 and (
            params["quota"] is None or params["quota"] <= 0 or params["quota"] > max_vfolder_size
        ):
            params["quota"] = max_vfolder_size

        # Prevent creation of vfolder with duplicated name.
        extra_vf_conds = [vfolders.c.name == params["name"]]
        if not unmanaged_path:
            extra_vf_conds.append(vfolders.c.host == folder_host)
        entries = await query_accessible_vfolders(
            conn,
            user_uuid,
            user_role=user_role,
            domain_name=domain_name,
            allowed_vfolder_types=allowed_vfolder_types,
            extra_vf_conds=(sa.and_(*extra_vf_conds)),
        )
        if len(entries) > 0:
            raise VFolderAlreadyExists

        # Check if group exists.
        if group_id_or_name and group_id is None:
            raise GroupNotFound
        if group_id is not None:
            if "group" not in allowed_vfolder_types:
                raise InvalidAPIParameters("group vfolder cannot be created in this host")
            if not request["is_admin"]:
                raise GenericForbidden("no permission")
            query = (
                sa.select([groups.c.id])
                .select_from(groups)
                .where(groups.c.domain_name == domain_name)
                .where(groups.c.id == group_id)
            )
            _gid = await conn.scalar(query)
            if str(_gid) != str(group_id):
                raise InvalidAPIParameters("No such group.")
        else:
            if "user" not in allowed_vfolder_types:
                raise InvalidAPIParameters("user vfolder cannot be created in this host")
        try:
            folder_id = uuid.uuid4()
            if not unmanaged_path:
                # Try to create actual only if vFolder is managed one
                async with root_ctx.storage_manager.request(
                    folder_host,
                    "POST",
                    "folder/create",
                    json={
                        "volume": root_ctx.storage_manager.split_host(folder_host)[1],
                        "vfid": str(folder_id),
                        "options": {"quota": params["quota"]},
                    },
                ):
                    pass
        except aiohttp.ClientResponseError:
            raise VFolderCreationFailed
        user_uuid = str(user_uuid) if group_id is None else None
        group_uuid = str(group_id) if group_id is not None else None
        ownership_type = "group" if group_uuid is not None else "user"
        insert_values = {
            "id": folder_id.hex,
            "name": params["name"],
            "usage_mode": params["usage_mode"],
            "permission": params["permission"],
            "last_used": None,
            "max_size": int(params["quota"] / (2**20)) if params["quota"] else None,  # in MBytes
            "host": folder_host,
            "creator": request["user"]["email"],
            "ownership_type": VFolderOwnershipType(ownership_type),
            "user": user_uuid,
            "group": group_uuid,
            "unmanaged_path": "",
            "cloneable": params["cloneable"],
            "status": VFolderOperationStatus.READY,
        }
        resp = {
            "id": folder_id.hex,
            "name": params["name"],
            "host": folder_host,
            "usage_mode": params["usage_mode"].value,
            "permission": params["permission"].value,
            "max_size": int(params["quota"] / (2**20)) if params["quota"] else None,  # in MBytes
            "creator": request["user"]["email"],
            "ownership_type": ownership_type,
            "user": user_uuid,
            "group": group_uuid,
            "cloneable": params["cloneable"],
            "status": VFolderOperationStatus.READY,
        }
        if unmanaged_path:
            insert_values.update(
                {
                    "host": "",
                    "unmanaged_path": unmanaged_path,
                }
            )
            resp["unmanaged_path"] = unmanaged_path
        query = sa.insert(vfolders, insert_values)
        try:
            result = await conn.execute(query)
        except sa.exc.DataError:
            raise InvalidAPIParameters
        assert result.rowcount == 1
    return web.json_response(resp, status=201)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("all", default=False): t.ToBool,
            tx.AliasedKey(["group_id", "groupId"], default=None): tx.UUID | t.String | t.Null,
            tx.AliasedKey(["owner_user_email", "ownerUserEmail"], default=None): t.Email | t.Null,
        }
    ),
)
async def list_folders(request: web.Request, params: Any) -> web.Response:
    resp = []
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    domain_name = request["user"]["domain_name"]

    log.info("VFOLDER.LIST (email:{}, ak:{})", request["user"]["email"], access_key)
    entries: List[Mapping[str, Any]] | Sequence[Mapping[str, Any]]
    owner_user_uuid, owner_user_role = await get_user_scopes(request, params)
    async with root_ctx.db.begin_readonly() as conn:
        allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
        if params["all"]:
            raise InvalidAPIParameters("Deprecated use of 'all' option")
        else:
            extra_vf_conds = None
            if params["group_id"] is not None:
                # Note: user folders should be returned even when group_id is specified.
                extra_vf_conds = (vfolders.c.group == params["group_id"]) | (
                    vfolders.c.user.isnot(None)
                )
            entries = await query_accessible_vfolders(
                conn,
                owner_user_uuid,
                user_role=owner_user_role,
                domain_name=domain_name,
                allowed_vfolder_types=allowed_vfolder_types,
                extra_vf_conds=extra_vf_conds,
            )
        for entry in entries:
            resp.append(
                {
                    "name": entry["name"],
                    "id": entry["id"].hex,
                    "host": entry["host"],
                    "status": entry["status"],
                    "usage_mode": entry["usage_mode"].value,
                    "created_at": str(entry["created_at"]),
                    "is_owner": entry["is_owner"],
                    "permission": entry["permission"].value,
                    "user": str(entry["user"]) if entry["user"] else None,
                    "group": str(entry["group"]) if entry["group"] else None,
                    "creator": entry["creator"],
                    "user_email": entry["user_email"],
                    "group_name": entry["group_name"],
                    "ownership_type": entry["ownership_type"].value,
                    "type": entry["ownership_type"].value,  # legacy
                    "cloneable": entry["cloneable"],
                    "max_files": entry["max_files"],
                    "max_size": entry["max_size"],
                    "cur_size": entry["cur_size"],
                }
            )
    return web.json_response(resp, status=200)


@superadmin_required
@server_status_required(ALL_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("id"): t.String,
        }
    ),
)
async def delete_by_id(request: web.Request, params: Any) -> web.Response:
    await ensure_vfolder_status(request, VFolderAccessStatus.DELETABLE, folder_id=params["id"])
    root_ctx: RootContext = request.app["_root.context"]
    app_ctx: PrivateContext = request.app["folders.context"]

    access_key = request["keypair"]["access_key"]
    user_uuid = request["user"]["uuid"]
    domain_name = request["user"]["domain_name"]
    resource_policy = request["keypair"]["resource_policy"]
    allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
    log.info(
        "VFOLDER.DELETE_BY_ID (email:{}, ak:{}, vf:{})",
        request["user"]["email"],
        access_key,
        params["id"],
    )
    async with root_ctx.db.begin() as conn:
        query = (
            sa.select([vfolders.c.host]).select_from(vfolders).where(vfolders.c.id == params["id"])
        )
        folder_host = await conn.scalar(query)
        await ensure_host_permission_allowed(
            conn,
            folder_host,
            allowed_vfolder_types=allowed_vfolder_types,
            user_uuid=user_uuid,
            resource_policy=resource_policy,
            domain_name=domain_name,
            permission=VFolderHostPermission.DELETE,
        )
    folder_id = uuid.UUID(params["id"])
    await initiate_vfolder_removal(
        root_ctx.db,
        [VFolderDeletionInfo(folder_id, folder_host)],
        root_ctx.storage_manager,
        app_ctx.storage_ptask_group,
    )
    return web.Response(status=204)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["group_id", "groupId"], default=None): tx.UUID | t.String | t.Null,
        }
    ),
)
async def list_hosts(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    log.info(
        "VFOLDER.LIST_HOSTS (emai:{}, ak:{})",
        request["user"]["email"],
        access_key,
    )
    domain_name = request["user"]["domain_name"]
    group_id = params["group_id"]
    domain_admin = request["user"]["role"] == UserRole.ADMIN
    resource_policy = request["keypair"]["resource_policy"]
    allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
    async with root_ctx.db.begin() as conn:
        allowed_hosts = VFolderHostPermissionMap()
        if "user" in allowed_vfolder_types:
            allowed_hosts_by_user = await get_allowed_vfolder_hosts_by_user(
                conn, resource_policy, domain_name, request["user"]["uuid"], group_id
            )
            allowed_hosts = allowed_hosts | allowed_hosts_by_user
        if "group" in allowed_vfolder_types:
            allowed_hosts_by_group = await get_allowed_vfolder_hosts_by_group(
                conn, resource_policy, domain_name, group_id, domain_admin=domain_admin
            )
            allowed_hosts = allowed_hosts | allowed_hosts_by_group
    all_volumes = await root_ctx.storage_manager.get_all_volumes()
    all_hosts = {f"{proxy_name}:{volume_data['name']}" for proxy_name, volume_data in all_volumes}
    allowed_hosts = VFolderHostPermissionMap(
        {host: perms for host, perms in allowed_hosts.items() if host in all_hosts}
    )
    default_host = await root_ctx.shared_config.get_raw("volumes/default_host")
    if default_host not in allowed_hosts:
        default_host = None
    volume_info = {
        f"{proxy_name}:{volume_data['name']}": {
            "backend": volume_data["backend"],
            "capabilities": volume_data["capabilities"],
        }
        for proxy_name, volume_data in all_volumes
        if f"{proxy_name}:{volume_data['name']}" in allowed_hosts
    }
    resp = {
        "default": default_host,
        "allowed": sorted(allowed_hosts),
        "volume_info": volume_info,
    }
    return web.json_response(resp, status=200)


@superadmin_required
@server_status_required(READ_ALLOWED)
async def list_all_hosts(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    log.info(
        "VFOLDER.LIST_ALL_HOSTS (email:{}, ak:{})",
        request["user"]["email"],
        access_key,
    )
    all_volumes = await root_ctx.storage_manager.get_all_volumes()
    all_hosts = {f"{proxy_name}:{volume_data['name']}" for proxy_name, volume_data in all_volumes}
    default_host = await root_ctx.shared_config.get_raw("volumes/default_host")
    if default_host not in all_hosts:
        default_host = None
    resp = {
        "default": default_host,
        "allowed": sorted(all_hosts),
    }
    return web.json_response(resp, status=200)


@superadmin_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("folder_host"): t.String,
        }
    )
)
async def get_volume_perf_metric(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    log.info(
        "VFOLDER.VOLUME_PERF_METRIC (email:{}, ak:{})",
        request["user"]["email"],
        access_key,
    )
    proxy_name, volume_name = root_ctx.storage_manager.split_host(params["folder_host"])
    async with root_ctx.storage_manager.request(
        proxy_name,
        "GET",
        "volume/performance-metric",
        json={
            "volume": volume_name,
        },
    ) as (_, storage_resp):
        storage_reply = await storage_resp.json()
    return web.json_response(storage_reply, status=200)


@auth_required
@server_status_required(READ_ALLOWED)
async def list_allowed_types(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    log.info(
        "VFOLDER.LIST_ALLOWED_TYPES (email:{}, ak:{})",
        request["user"]["email"],
        access_key,
    )
    allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
    return web.json_response(allowed_vfolder_types, status=200)


@auth_required
@server_status_required(READ_ALLOWED)
@vfolder_permission_required(VFolderPermission.READ_ONLY)
async def get_info(request: web.Request, row: VFolderRow) -> web.Response:
    await ensure_vfolder_status(
        request, VFolderAccessStatus.READABLE, folder_name=request.match_info["name"]
    )
    root_ctx: RootContext = request.app["_root.context"]
    resp: Dict[str, Any] = {}
    folder_name = request.match_info["name"]
    access_key = request["keypair"]["access_key"]
    log.info(
        "VFOLDER.GETINFO (email:{}, ak:{}, vf:{})",
        request["user"]["email"],
        access_key,
        folder_name,
    )
    if row["permission"] is None:
        is_owner = True
        permission = VFolderPermission.OWNER_PERM
    else:
        is_owner = row["is_owner"]
        permission = row["permission"]
    proxy_name, volume_name = root_ctx.storage_manager.split_host(row["host"])
    async with root_ctx.storage_manager.request(
        proxy_name,
        "GET",
        "folder/usage",
        json={
            "volume": volume_name,
            "vfid": str(row["id"]),
        },
    ) as (_, storage_resp):
        usage = await storage_resp.json()
    resp = {
        "name": row["name"],
        "id": row["id"].hex,
        "host": row["host"],
        "status": row["status"],
        "numFiles": usage["file_count"],  # legacy
        "num_files": usage["file_count"],
        "used_bytes": usage["used_bytes"],  # added in v20.09
        "created": str(row["created_at"]),  # legacy
        "created_at": str(row["created_at"]),
        "last_used": str(row["created_at"]),
        "user": str(row["user"]),
        "group": str(row["group"]),
        "type": "user" if row["user"] is not None else "group",
        "is_owner": is_owner,
        "permission": permission,
        "usage_mode": row["usage_mode"],
        "cloneable": row["cloneable"],
        "max_size": row["max_size"],
        "cur_size": row["cur_size"],
    }
    return web.json_response(resp, status=200)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("folder_host"): t.String,
            t.Key("id"): tx.UUID,
        }
    )
)
async def get_quota(request: web.Request, params: Any) -> web.Response:
    await ensure_vfolder_status(request, VFolderAccessStatus.READABLE, folder_id=params["id"])
    root_ctx: RootContext = request.app["_root.context"]
    proxy_name, volume_name = root_ctx.storage_manager.split_host(params["folder_host"])
    log.info(
        "VFOLDER.GET_QUOTA (email:{}, volume_name:{}, vf:{})",
        request["user"]["email"],
        volume_name,
        params["id"],
    )

    # Permission check for the requested vfolder.
    user_role = request["user"]["role"]
    user_uuid = request["user"]["uuid"]
    domain_name = request["user"]["domain_name"]
    if user_role == UserRole.SUPERADMIN:
        pass
    else:
        allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
        async with root_ctx.db.begin_readonly() as conn:
            extra_vf_conds = [vfolders.c.id == params["id"]]
            entries = await query_accessible_vfolders(
                conn,
                user_uuid,
                user_role=user_role,
                domain_name=domain_name,
                allowed_vfolder_types=allowed_vfolder_types,
                extra_vf_conds=(sa.and_(*extra_vf_conds)),
            )
        if len(entries) < 0:
            raise VFolderNotFound("no such accessible vfolder")

    async with root_ctx.storage_manager.request(
        proxy_name,
        "GET",
        "volume/quota",
        json={
            "volume": volume_name,
            "vfid": str(params["id"]),
        },
    ) as (_, storage_resp):
        storage_reply = await storage_resp.json()
    return web.json_response(storage_reply, status=200)


@auth_required
@server_status_required(ALL_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("folder_host"): t.String,
            t.Key("id"): tx.UUID,
            t.Key("input"): t.Mapping(t.String, t.Any),
        }
    ),
)
async def update_quota(request: web.Request, params: Any) -> web.Response:
    await ensure_vfolder_status(request, VFolderAccessStatus.UPDATABLE, folder_id=params["id"])
    root_ctx: RootContext = request.app["_root.context"]
    folder_host = params["folder_host"]
    proxy_name, volume_name = root_ctx.storage_manager.split_host(folder_host)
    quota = int(params["input"]["size_bytes"])
    log.info(
        "VFOLDER.UPDATE_QUOTA (email:{}, volume_name:{}, quota:{}, vf:{})",
        request["user"]["email"],
        volume_name,
        quota,
        params["id"],
    )

    # Permission check for the requested vfolder.
    user_role = request["user"]["role"]
    user_uuid = request["user"]["uuid"]
    domain_name = request["user"]["domain_name"]
    resource_policy = request["keypair"]["resource_policy"]

    if user_role == UserRole.SUPERADMIN:
        pass
    else:
        allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
        async with root_ctx.db.begin_readonly() as conn:
            await ensure_host_permission_allowed(
                conn,
                folder_host,
                allowed_vfolder_types=allowed_vfolder_types,
                user_uuid=user_uuid,
                resource_policy=resource_policy,
                domain_name=domain_name,
                permission=VFolderHostPermission.MODIFY,
            )
            extra_vf_conds = [vfolders.c.id == params["id"]]
            entries = await query_accessible_vfolders(
                conn,
                user_uuid,
                user_role=user_role,
                domain_name=domain_name,
                allowed_vfolder_types=allowed_vfolder_types,
                extra_vf_conds=(sa.and_(*extra_vf_conds)),
            )
        if len(entries) < 0:
            raise VFolderNotFound("no such accessible vfolder")

    # Limit vfolder size quota if it is larger than max_vfolder_size of the resource policy.
    max_vfolder_size = resource_policy.get("max_vfolder_size", 0)
    if max_vfolder_size > 0 and (quota <= 0 or quota > max_vfolder_size):
        quota = max_vfolder_size

    async with root_ctx.storage_manager.request(
        proxy_name,
        "PATCH",
        "volume/quota",
        json={
            "volume": volume_name,
            "vfid": str(params["id"]),
            "size_bytes": quota,
        },
    ):
        pass

    # Update the quota for the vfolder in DB.
    async with root_ctx.db.begin() as conn:
        query = (
            sa.update(vfolders)
            .values(max_size=math.ceil(quota / 2**20))  # in Mbytes
            .where(vfolders.c.id == params["id"])
        )
        result = await conn.execute(query)
        assert result.rowcount == 1

    return web.json_response({"size_bytes": quota}, status=200)


@superadmin_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("folder_host"): t.String,
            t.Key("id"): tx.UUID,
        }
    )
)
async def get_usage(request: web.Request, params: Any) -> web.Response:
    await ensure_vfolder_status(request, VFolderAccessStatus.READABLE, folder_id=params["id"])
    root_ctx: RootContext = request.app["_root.context"]
    proxy_name, volume_name = root_ctx.storage_manager.split_host(params["folder_host"])
    log.info(
        "VFOLDER.GET_USAGE (email:{}, volume_name:{}, vf:{})",
        request["user"]["email"],
        volume_name,
        params["id"],
    )
    async with root_ctx.storage_manager.request(
        proxy_name,
        "GET",
        "folder/usage",
        json={
            "volume": volume_name,
            "vfid": str(params["id"]),
        },
    ) as (_, storage_resp):
        usage = await storage_resp.json()
    return web.json_response(usage, status=200)


@superadmin_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("folder_host"): t.String,
            t.Key("id"): tx.UUID,
        }
    )
)
async def get_used_bytes(request: web.Request, params: Any) -> web.Response:
    await ensure_vfolder_status(request, VFolderAccessStatus.READABLE, folder_id=params["id"])
    root_ctx: RootContext = request.app["_root.context"]
    proxy_name, volume_name = root_ctx.storage_manager.split_host(params["folder_host"])
    log.info("VFOLDER.GET_USED_BYTES (volume_name:{}, vf:{})", volume_name, params["id"])
    async with root_ctx.storage_manager.request(
        proxy_name,
        "GET",
        "folder/used-bytes",
        json={
            "volume": volume_name,
            "vfid": str(params["id"]),
        },
    ) as (_, storage_resp):
        usage = await storage_resp.json()
    return web.json_response(usage, status=200)


@auth_required
@server_status_required(ALL_ALLOWED)
@vfolder_permission_required(VFolderPermission.OWNER_PERM)
@check_api_params(
    t.Dict(
        {
            t.Key("new_name"): tx.Slug(allow_dot=True),
        }
    )
)
async def rename_vfolder(request: web.Request, params: Any, row: VFolderRow) -> web.Response:
    await ensure_vfolder_status(
        request, VFolderAccessStatus.UPDATABLE, folder_name=request.match_info["name"]
    )
    root_ctx: RootContext = request.app["_root.context"]
    old_name = request.match_info["name"]
    access_key = request["keypair"]["access_key"]
    domain_name = request["user"]["domain_name"]
    user_role = request["user"]["role"]
    user_uuid = request["user"]["uuid"]
    resource_policy = request["keypair"]["resource_policy"]
    new_name = params["new_name"]
    allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
    log.info(
        "VFOLDER.RENAME (email:{}, ak:{}, vf.old:{}, vf.new:{})",
        request["user"]["email"],
        access_key,
        old_name,
        new_name,
    )
    async with root_ctx.db.begin() as conn:
        entries = await query_accessible_vfolders(
            conn,
            user_uuid,
            user_role=user_role,
            domain_name=domain_name,
            allowed_vfolder_types=allowed_vfolder_types,
        )
        for entry in entries:
            if entry["name"] == new_name:
                raise InvalidAPIParameters(
                    "One of your accessible vfolders already has " "the name you requested."
                )
        for entry in entries:
            if entry["name"] == old_name:
                if not entry["is_owner"]:
                    raise InvalidAPIParameters(
                        "Cannot change the name of a vfolder " "that is not owned by myself."
                    )
                await ensure_host_permission_allowed(
                    conn,
                    entry["host"],
                    allowed_vfolder_types=allowed_vfolder_types,
                    user_uuid=user_uuid,
                    resource_policy=resource_policy,
                    domain_name=domain_name,
                    permission=VFolderHostPermission.MODIFY,
                )
                query = (
                    sa.update(vfolders).values(name=new_name).where(vfolders.c.id == entry["id"])
                )
                await conn.execute(query)
                break
    return web.Response(status=201)


@auth_required
@server_status_required(ALL_ALLOWED)
@vfolder_permission_required(VFolderPermission.OWNER_PERM)
@check_api_params(
    t.Dict(
        {
            t.Key("cloneable", default=None): t.Bool | t.Null,
            t.Key("permission", default=None): tx.Enum(VFolderPermission) | t.Null,
        }
    )
)
async def update_vfolder_options(
    request: web.Request, params: Any, row: VFolderRow
) -> web.Response:
    await ensure_vfolder_status(
        request, VFolderAccessStatus.UPDATABLE, folder_name=request.match_info["name"]
    )
    root_ctx: RootContext = request.app["_root.context"]
    user_uuid = request["user"]["uuid"]
    domain_name = request["user"]["domain_name"]
    resource_policy = request["keypair"]["resource_policy"]
    allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
    async with root_ctx.db.begin_readonly() as conn:
        query = sa.select([vfolders.c.host]).select_from(vfolders).where(vfolders.c.id == row["id"])
        folder_host = await conn.scalar(query)
        await ensure_host_permission_allowed(
            conn,
            folder_host,
            allowed_vfolder_types=allowed_vfolder_types,
            user_uuid=user_uuid,
            resource_policy=resource_policy,
            domain_name=domain_name,
            permission=VFolderHostPermission.MODIFY,
        )

    updated_fields = {}
    if params["cloneable"] is not None and params["cloneable"] != row["cloneable"]:
        updated_fields["cloneable"] = params["cloneable"]
    if params["permission"] is not None and params["permission"] != row["permission"]:
        updated_fields["permission"] = params["permission"]
    if not row["is_owner"]:
        raise InvalidAPIParameters(
            "Cannot change the options of a vfolder " "that is not owned by myself."
        )

    if len(updated_fields) > 0:
        async with root_ctx.db.begin() as conn:
            query = sa.update(vfolders).values(**updated_fields).where(vfolders.c.id == row["id"])
            await conn.execute(query)
    return web.Response(status=201)


@auth_required
@server_status_required(READ_ALLOWED)
@vfolder_permission_required(VFolderPermission.READ_WRITE)
@check_api_params(
    t.Dict(
        {
            t.Key("path"): t.String,
            t.Key("parents", default=True): t.ToBool,
            t.Key("exist_ok", default=False): t.ToBool,
        }
    )
)
async def mkdir(request: web.Request, params: Any, row: VFolderRow) -> web.Response:
    await ensure_vfolder_status(
        request, VFolderAccessStatus.UPDATABLE, folder_name=request.match_info["name"]
    )
    root_ctx: RootContext = request.app["_root.context"]
    folder_name = request.match_info["name"]
    access_key = request["keypair"]["access_key"]
    log.info(
        "VFOLDER.MKDIR (email:{}, ak:{}, vf:{}, path:{})",
        request["user"]["email"],
        access_key,
        folder_name,
        params["path"],
    )
    proxy_name, volume_name = root_ctx.storage_manager.split_host(row["host"])
    async with root_ctx.storage_manager.request(
        proxy_name,
        "POST",
        "folder/file/mkdir",
        json={
            "volume": volume_name,
            "vfid": str(row["id"]),
            "relpath": params["path"],
            "parents": params["parents"],
            "exist_ok": params["exist_ok"],
        },
    ):
        pass
    return web.Response(status=201)


@auth_required
@server_status_required(READ_ALLOWED)
@vfolder_permission_required(VFolderPermission.READ_ONLY)
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["path", "file"]): t.String,
            t.Key("archive", default=False): t.ToBool,
        }
    )
)
async def create_download_session(
    request: web.Request, params: Any, row: VFolderRow
) -> web.Response:
    await ensure_vfolder_status(
        request, VFolderAccessStatus.UPDATABLE, folder_name=request.match_info["name"]
    )
    root_ctx: RootContext = request.app["_root.context"]
    log_fmt = "VFOLDER.CREATE_DOWNLOAD_SESSION(email:{}, ak:{}, vf:{}, path:{})"
    log_args = (
        request["user"]["email"],
        request["keypair"]["access_key"],
        row["name"],
        params["path"],
    )
    log.info(log_fmt, *log_args)
    unmanaged_path = row["unmanaged_path"]
    user_uuid = request["user"]["uuid"]
    folder_host = row["host"]
    domain_name = request["user"]["domain_name"]
    resource_policy = request["keypair"]["resource_policy"]
    allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
    async with root_ctx.db.begin_readonly() as conn:
        await ensure_host_permission_allowed(
            conn,
            folder_host,
            allowed_vfolder_types=allowed_vfolder_types,
            user_uuid=user_uuid,
            resource_policy=resource_policy,
            domain_name=domain_name,
            permission=VFolderHostPermission.DOWNLOAD_FILE,
        )
    proxy_name, volume_name = root_ctx.storage_manager.split_host(folder_host)
    async with root_ctx.storage_manager.request(
        proxy_name,
        "POST",
        "folder/file/download",
        json={
            "volume": volume_name,
            "vfid": str(row["id"]),
            "relpath": params["path"],
            "archive": params["archive"],
            "unmanaged_path": unmanaged_path if unmanaged_path else None,
        },
    ) as (client_api_url, storage_resp):
        storage_reply = await storage_resp.json()
        resp = {
            "token": storage_reply["token"],
            "url": str(client_api_url / "download"),
        }
    return web.json_response(resp, status=200)


@auth_required
@server_status_required(READ_ALLOWED)
@vfolder_permission_required(VFolderPermission.READ_WRITE)
@check_api_params(
    t.Dict(
        {
            t.Key("path"): t.String,
            t.Key("size"): t.ToInt,
        }
    )
)
async def create_upload_session(request: web.Request, params: Any, row: VFolderRow) -> web.Response:
    await ensure_vfolder_status(
        request, VFolderAccessStatus.UPDATABLE, folder_name=request.match_info["name"]
    )
    root_ctx: RootContext = request.app["_root.context"]
    folder_name = request.match_info["name"]
    access_key = request["keypair"]["access_key"]
    log_fmt = "VFOLDER.CREATE_UPLOAD_SESSION (email:{}, ak:{}, vf:{}, path:{})"
    log_args = (request["user"]["email"], access_key, folder_name, params["path"])
    log.info(log_fmt, *log_args)
    user_uuid = request["user"]["uuid"]
    domain_name = request["user"]["domain_name"]
    folder_host = row["host"]
    resource_policy = request["keypair"]["resource_policy"]
    allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
    async with root_ctx.db.begin_readonly() as conn:
        await ensure_host_permission_allowed(
            conn,
            folder_host,
            allowed_vfolder_types=allowed_vfolder_types,
            user_uuid=user_uuid,
            resource_policy=resource_policy,
            domain_name=domain_name,
            permission=VFolderHostPermission.UPLOAD_FILE,
        )
    proxy_name, volume_name = root_ctx.storage_manager.split_host(folder_host)
    async with root_ctx.storage_manager.request(
        proxy_name,
        "POST",
        "folder/file/upload",
        json={
            "volume": volume_name,
            "vfid": str(row["id"]),
            "relpath": params["path"],
            "size": params["size"],
        },
    ) as (client_api_url, storage_resp):
        storage_reply = await storage_resp.json()
        resp = {
            "token": storage_reply["token"],
            "url": str(client_api_url / "upload"),
        }
    return web.json_response(resp, status=200)


@auth_required
@server_status_required(READ_ALLOWED)
@vfolder_permission_required(VFolderPermission.READ_WRITE)
@check_api_params(
    t.Dict(
        {
            t.Key("target_path"): t.String,
            t.Key("new_name"): t.String,
            t.Key("is_dir", default=False): t.ToBool,  # ignored since 22.03
        }
    )
)
async def rename_file(request: web.Request, params: Any, row: VFolderRow) -> web.Response:
    await ensure_vfolder_status(
        request, VFolderAccessStatus.UPDATABLE, folder_name=request.match_info["name"]
    )
    root_ctx: RootContext = request.app["_root.context"]
    folder_name = request.match_info["name"]
    access_key = request["keypair"]["access_key"]
    user_uuid = request["user"]["uuid"]
    domain_name = request["user"]["domain_name"]
    folder_host = row["host"]
    resource_policy = request["keypair"]["resource_policy"]
    allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
    async with root_ctx.db.begin_readonly() as conn:
        await ensure_host_permission_allowed(
            conn,
            folder_host,
            allowed_vfolder_types=allowed_vfolder_types,
            user_uuid=user_uuid,
            resource_policy=resource_policy,
            domain_name=domain_name,
            permission=VFolderHostPermission.MODIFY,
        )
    log.info(
        "VFOLDER.RENAME_FILE (email:{}, ak:{}, vf:{}, target_path:{}, new_name:{})",
        request["user"]["email"],
        access_key,
        folder_name,
        params["target_path"],
        params["new_name"],
    )
    proxy_name, volume_name = root_ctx.storage_manager.split_host(folder_host)
    async with root_ctx.storage_manager.request(
        proxy_name,
        "POST",
        "folder/file/rename",
        json={
            "volume": volume_name,
            "vfid": str(row["id"]),
            "relpath": params["target_path"],
            "new_name": params["new_name"],
        },
    ):
        pass
    return web.json_response({}, status=200)


@auth_required
@server_status_required(READ_ALLOWED)
@vfolder_permission_required(VFolderPermission.READ_WRITE)
@check_api_params(
    t.Dict(
        {
            t.Key("src"): t.String,
            t.Key("dst"): t.String,
        }
    )
)
async def move_file(request: web.Request, params: Any, row: VFolderRow) -> web.Response:
    await ensure_vfolder_status(
        request, VFolderAccessStatus.UPDATABLE, folder_name=request.match_info["name"]
    )
    root_ctx: RootContext = request.app["_root.context"]
    folder_name = request.match_info["name"]
    access_key = request["keypair"]["access_key"]
    log.info(
        "VFOLDER.MOVE_FILE (email:{}, ak:{}, vf:{}, src:{}, dst:{})",
        request["user"]["email"],
        access_key,
        folder_name,
        params["src"],
        params["dst"],
    )
    proxy_name, volume_name = root_ctx.storage_manager.split_host(row["host"])
    async with root_ctx.storage_manager.request(
        proxy_name,
        "POST",
        "folder/file/move",
        json={
            "volume": volume_name,
            "vfid": str(row["id"]),
            "src_relpath": params["src"],
            "dst_relpath": params["dst"],
        },
    ):
        pass
    return web.json_response({}, status=200)


@auth_required
@server_status_required(READ_ALLOWED)
@vfolder_permission_required(VFolderPermission.READ_WRITE)
@check_api_params(
    t.Dict(
        {
            t.Key("files"): t.List(t.String),
            t.Key("recursive", default=False): t.ToBool,
        }
    )
)
async def delete_files(request: web.Request, params: Any, row: VFolderRow) -> web.Response:
    await ensure_vfolder_status(
        request, VFolderAccessStatus.UPDATABLE, folder_name=request.match_info["name"]
    )
    root_ctx: RootContext = request.app["_root.context"]
    folder_name = request.match_info["name"]
    access_key = request["keypair"]["access_key"]
    recursive = params["recursive"]
    log.info(
        "VFOLDER.DELETE_FILES (email:{}, ak:{}, vf:{}, path:{}, recursive:{})",
        request["user"]["email"],
        access_key,
        folder_name,
        params["files"],
        recursive,
    )
    proxy_name, volume_name = root_ctx.storage_manager.split_host(row["host"])
    async with root_ctx.storage_manager.request(
        proxy_name,
        "POST",
        "folder/file/delete",
        json={
            "volume": volume_name,
            "vfid": str(row["id"]),
            "relpaths": params["files"],
            "recursive": recursive,
        },
    ):
        pass
    return web.json_response({}, status=200)


@auth_required
@server_status_required(READ_ALLOWED)
@vfolder_permission_required(VFolderPermission.READ_ONLY)
@check_api_params(
    t.Dict(
        {
            t.Key("path", default=""): t.String(allow_blank=True),
        }
    )
)
async def list_files(request: web.Request, params: Any, row: VFolderRow) -> web.Response:
    await ensure_vfolder_status(
        request, VFolderAccessStatus.READABLE, folder_name=request.match_info["name"]
    )
    root_ctx: RootContext = request.app["_root.context"]
    folder_name = request.match_info["name"]
    access_key = request["keypair"]["access_key"]
    log.info(
        "VFOLDER.LIST_FILES (email:{}, ak:{}, vf:{}, path:{})",
        request["user"]["email"],
        access_key,
        folder_name,
        params["path"],
    )
    proxy_name, volume_name = root_ctx.storage_manager.split_host(row["host"])
    async with root_ctx.storage_manager.request(
        proxy_name,
        "POST",
        "folder/file/list",
        json={
            "volume": volume_name,
            "vfid": str(row["id"]),
            "relpath": params["path"],
        },
    ) as (_, storage_resp):
        result = await storage_resp.json()
        resp = {
            "items": [
                {
                    "name": item["name"],
                    "type": item["type"],
                    "size": item["stat"]["size"],  # humanize?
                    "mode": oct(item["stat"]["mode"])[2:][-3:],
                    "created": item["stat"]["created"],
                    "modified": item["stat"]["modified"],
                }
                for item in result["items"]
            ],
            "files": json.dumps(
                [  # for legacy (to be removed in 21.03)
                    {
                        "filename": item["name"],
                        "size": item["stat"]["size"],
                        "mode": stat.filemode(item["stat"]["mode"]),
                        "ctime": datetime.fromisoformat(item["stat"]["created"]).timestamp(),
                        "atime": 0,
                        "mtime": datetime.fromisoformat(item["stat"]["modified"]).timestamp(),
                    }
                    for item in result["items"]
                ]
            ),
        }
    return web.json_response(resp, status=200)


@auth_required
@server_status_required(READ_ALLOWED)
async def list_sent_invitations(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    log.info(
        "VFOLDER.LIST_SENT_INVITATIONS (email:{}, ak:{})",
        request["user"]["email"],
        access_key,
    )
    async with root_ctx.db.begin() as conn:
        j = sa.join(vfolders, vfolder_invitations, vfolders.c.id == vfolder_invitations.c.vfolder)
        query = (
            sa.select([vfolder_invitations, vfolders.c.name])
            .select_from(j)
            .where(
                (vfolder_invitations.c.inviter == request["user"]["email"])
                & (vfolder_invitations.c.state == VFolderInvitationState.PENDING),
            )
        )
        result = await conn.execute(query)
        invitations = result.fetchall()
    invs_info = []
    for inv in invitations:
        invs_info.append(
            {
                "id": str(inv.id),
                "inviter": inv.inviter,
                "invitee": inv.invitee,
                "perm": inv.permission,
                "state": inv.state.value,
                "created_at": str(inv.created_at),
                "modified_at": str(inv.modified_at),
                "vfolder_id": str(inv.vfolder),
                "vfolder_name": inv.name,
            }
        )
    resp = {"invitations": invs_info}
    return web.json_response(resp, status=200)


@auth_required
@server_status_required(ALL_ALLOWED)
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["perm", "permission"]): VFolderPermissionValidator,
        }
    ),
)
async def update_invitation(request: web.Request, params: Any) -> web.Response:
    """
    Update sent invitation's permission. Other fields are not allowed to be updated.
    """
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    inv_id = request.match_info["inv_id"]
    perm = params["perm"]
    log.info(
        "VFOLDER.UPDATE_INVITATION (email:{}, ak:{}, inv:{})",
        request["user"]["email"],
        access_key,
        inv_id,
    )
    async with root_ctx.db.begin() as conn:
        query = (
            sa.update(vfolder_invitations)
            .values(permission=perm)
            .where(
                (vfolder_invitations.c.id == inv_id)
                & (vfolder_invitations.c.inviter == request["user"]["email"])
                & (vfolder_invitations.c.state == VFolderInvitationState.PENDING),
            )
        )
        await conn.execute(query)
    resp = {"msg": f"vfolder invitation updated: {inv_id}."}
    return web.json_response(resp, status=200)


@auth_required
@server_status_required(ALL_ALLOWED)
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["perm", "permission"], default="rw"): VFolderPermissionValidator,
            tx.AliasedKey(["emails", "user_ids", "userIDs"]): t.List(t.String),
        }
    ),
)
async def invite(request: web.Request, params: Any) -> web.Response:
    await ensure_vfolder_status(
        request, VFolderAccessStatus.UPDATABLE, folder_name=request.match_info["name"]
    )
    root_ctx: RootContext = request.app["_root.context"]
    folder_name = request.match_info["name"]
    access_key = request["keypair"]["access_key"]
    user_uuid = request["user"]["uuid"]
    perm = params["perm"]
    invitee_emails = params["emails"]
    log.info(
        "VFOLDER.INVITE (email:{}, ak:{}, vf:{}, inv.users:{})",
        request["user"]["email"],
        access_key,
        folder_name,
        ",".join(invitee_emails),
    )
    domain_name = request["user"]["domain_name"]
    resource_policy = request["keypair"]["resource_policy"]
    if folder_name.startswith("."):
        raise GenericForbidden("Cannot share private dot-prefixed vfolders.")
    async with root_ctx.db.begin_readonly() as conn:
        # Get virtual folder.
        query = (
            sa.select("*")
            .select_from(vfolders)
            .where((vfolders.c.user == user_uuid) & (vfolders.c.name == folder_name))
        )
        try:
            result = await conn.execute(query)
        except sa.exc.DataError:
            raise InvalidAPIParameters
        vf = result.first()
        if vf is None:
            raise VFolderNotFound()
        folder_host = vf.host
        allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
        await ensure_host_permission_allowed(
            conn,
            folder_host,
            allowed_vfolder_types=allowed_vfolder_types,
            user_uuid=user_uuid,
            resource_policy=resource_policy,
            domain_name=domain_name,
            permission=VFolderHostPermission.INVITE_OTHERS,
        )
    async with root_ctx.db.begin() as conn:
        # Get invited user's keypairs except vfolder owner.
        query = (
            sa.select([keypairs.c.user_id, keypairs.c.user])
            .select_from(keypairs)
            .where(keypairs.c.user_id.in_(invitee_emails))
            .where(keypairs.c.user_id != request["user"]["email"])
        )
        try:
            result = await conn.execute(query)
        except sa.exc.DataError:
            raise InvalidAPIParameters
        kps = result.fetchall()
        if len(kps) < 1:
            raise ObjectNotFound(object_name="vfolder invitation")

        # Prevent inviting user who already share the target folder.
        invitee_uuids = [kp.user for kp in kps]
        j = sa.join(vfolders, vfolder_permissions, vfolders.c.id == vfolder_permissions.c.vfolder)
        query = (
            sa.select([sa.func.count()])
            .select_from(j)
            .where(
                (vfolders.c.user.in_(invitee_uuids) | vfolder_permissions.c.user.in_(invitee_uuids))
                & (vfolders.c.name == folder_name),
            )
        )
        result = await conn.execute(query)
        if result.scalar() > 0:
            raise VFolderAlreadyExists

        # Create invitation.
        invitees = [kp.user_id for kp in kps]
        invited_ids = []
        for invitee in set(invitees):
            inviter = request["user"]["id"]
            # Do not create invitation if already exists.
            query = (
                sa.select([sa.func.count()])
                .select_from(vfolder_invitations)
                .where(
                    (vfolder_invitations.c.inviter == inviter)
                    & (vfolder_invitations.c.invitee == invitee)
                    & (vfolder_invitations.c.vfolder == vf.id)
                    & (vfolder_invitations.c.state == VFolderInvitationState.PENDING),
                )
            )
            result = await conn.execute(query)
            if result.scalar() > 0:
                continue

            # TODO: insert multiple values with one query.
            #       insert().values([{}, {}, ...]) does not work:
            #       sqlalchemy.exc.CompileError: The 'default' dialect with current
            #       database version settings does not support in-place multirow
            #       inserts.
            query = sa.insert(
                vfolder_invitations,
                {
                    "id": uuid.uuid4().hex,
                    "permission": perm,
                    "vfolder": vf.id,
                    "inviter": inviter,
                    "invitee": invitee,
                    "state": VFolderInvitationState.PENDING,
                },
            )
            try:
                await conn.execute(query)
                invited_ids.append(invitee)
            except sa.exc.DataError:
                pass
    resp = {"invited_ids": invited_ids}
    return web.json_response(resp, status=201)


@auth_required
@server_status_required(READ_ALLOWED)
async def invitations(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    log.info(
        "VFOLDER.INVITATIONS (email:{}, ak:{})",
        request["user"]["email"],
        access_key,
    )
    async with root_ctx.db.begin() as conn:
        j = sa.join(vfolders, vfolder_invitations, vfolders.c.id == vfolder_invitations.c.vfolder)
        query = (
            sa.select([vfolder_invitations, vfolders.c.name])
            .select_from(j)
            .where(
                (vfolder_invitations.c.invitee == request["user"]["id"])
                & (vfolder_invitations.c.state == VFolderInvitationState.PENDING),
            )
        )
        result = await conn.execute(query)
        invitations = result.fetchall()
    invs_info = []
    for inv in invitations:
        invs_info.append(
            {
                "id": str(inv.id),
                "inviter": inv.inviter,
                "invitee": inv.invitee,
                "perm": inv.permission,
                "state": inv.state,
                "created_at": str(inv.created_at),
                "modified_at": str(inv.modified_at),
                "vfolder_id": str(inv.vfolder),
                "vfolder_name": inv.name,
            }
        )
    resp = {"invitations": invs_info}
    return web.json_response(resp, status=200)


@auth_required
@server_status_required(ALL_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("inv_id"): t.String,
        }
    ),
)
async def accept_invitation(request: web.Request, params: Any) -> web.Response:
    """Accept invitation by invitee.

    * `inv_ak` parameter is removed from 19.06 since virtual folder's ownership is
    moved from keypair to a user or a group.

    :param inv_id: ID of vfolder_invitations row.
    """
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    user_uuid = request["user"]["uuid"]
    inv_id = params["inv_id"]
    log.info(
        "VFOLDER.ACCEPT_INVITATION (email:{}, ak:{}, inv:{})",
        request["user"]["email"],
        access_key,
        inv_id,
    )
    async with root_ctx.db.begin() as conn:
        # Get invitation.
        query = (
            sa.select([vfolder_invitations])
            .select_from(vfolder_invitations)
            .where(
                (vfolder_invitations.c.id == inv_id)
                & (vfolder_invitations.c.state == VFolderInvitationState.PENDING),
            )
        )
        result = await conn.execute(query)
        invitation = result.first()
        if invitation is None:
            raise ObjectNotFound(object_name="vfolder invitation")

        # Get target virtual folder.
        query = (
            sa.select([vfolders.c.name])
            .select_from(vfolders)
            .where(vfolders.c.id == invitation.vfolder)
        )
        result = await conn.execute(query)
        target_vfolder = result.first()
        if target_vfolder is None:
            raise VFolderNotFound

        # Prevent accepting vfolder with duplicated name.
        j = sa.join(
            vfolders,
            vfolder_permissions,
            vfolders.c.id == vfolder_permissions.c.vfolder,
            isouter=True,
        )
        query = (
            sa.select([sa.func.count()])
            .select_from(j)
            .where(
                ((vfolders.c.user == user_uuid) | (vfolder_permissions.c.user == user_uuid))
                & (vfolders.c.name == target_vfolder.name),
            )
        )
        result = await conn.execute(query)
        if result.scalar() > 0:
            raise VFolderAlreadyExists

        # Create permission relation between the vfolder and the invitee.
        query = sa.insert(
            vfolder_permissions,
            {
                "permission": VFolderPermission(invitation.permission),
                "vfolder": invitation.vfolder,
                "user": user_uuid,
            },
        )
        await conn.execute(query)

        # Clear used invitation.
        query = (
            sa.update(vfolder_invitations)
            .where(vfolder_invitations.c.id == inv_id)
            .values(state=VFolderInvitationState.ACCEPTED)
        )
        await conn.execute(query)
    return web.json_response({})


@auth_required
@server_status_required(ALL_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("inv_id"): t.String,
        }
    )
)
async def delete_invitation(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    request_email = request["user"]["email"]
    inv_id = params["inv_id"]
    log.info(
        "VFOLDER.DELETE_INVITATION (email:{}, ak:{}, inv:{})",
        request["user"]["email"],
        access_key,
        inv_id,
    )
    try:
        async with root_ctx.db.begin() as conn:
            query = (
                sa.select(
                    [
                        vfolder_invitations.c.inviter,
                        vfolder_invitations.c.invitee,
                    ]
                )
                .select_from(vfolder_invitations)
                .where(
                    (vfolder_invitations.c.id == inv_id)
                    & (vfolder_invitations.c.state == VFolderInvitationState.PENDING),
                )
            )
            result = await conn.execute(query)
            row = result.first()
            if row is None:
                raise ObjectNotFound(object_name="vfolder invitation")
            if request_email == row.inviter:
                state = VFolderInvitationState.CANCELED
            elif request_email == row.invitee:
                state = VFolderInvitationState.REJECTED
            else:
                raise GenericForbidden("Cannot change other user's invitaiton")
            query = (
                sa.update(vfolder_invitations)
                .values(state=state)
                .where(vfolder_invitations.c.id == inv_id)
            )
            await conn.execute(query)
    except sa.exc.IntegrityError as e:
        raise InternalServerError(f"integrity error: {e}")
    except (asyncio.CancelledError, asyncio.TimeoutError):
        raise
    except Exception as e:
        raise InternalServerError(f"unexpected error: {e}")
    return web.json_response({})


@admin_required
@server_status_required(ALL_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("permission", default="rw"): VFolderPermissionValidator,
            t.Key("emails"): t.List(t.String),
        }
    ),
)
async def share(request: web.Request, params: Any) -> web.Response:
    await ensure_vfolder_status(
        request, VFolderAccessStatus.UPDATABLE, folder_name=request.match_info["name"]
    )
    """
    Share a group folder to users with overriding permission.

    This will create vfolder_permission(s) relation directly without
    creating invitation(s). Only group-type vfolders are allowed to
    be shared directly.
    """
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    folder_name = request.match_info["name"]
    log.info(
        "VFOLDER.SHARE (email:{}, ak:{}, vf:{}, perm:{}, users:{})",
        request["user"]["email"],
        access_key,
        folder_name,
        params["permission"],
        ",".join(params["emails"]),
    )
    user_uuid = request["user"]["uuid"]
    domain_name = request["user"]["domain_name"]
    resource_policy = request["keypair"]["resource_policy"]
    async with root_ctx.db.begin() as conn:
        from ..models import association_groups_users as agus

        # Get the group-type virtual folder.
        query = (
            sa.select([vfolders.c.id, vfolders.c.host, vfolders.c.ownership_type, vfolders.c.group])
            .select_from(vfolders)
            .where(
                (vfolders.c.ownership_type == VFolderOwnershipType.GROUP)
                & (vfolders.c.name == folder_name),
            )
        )
        result = await conn.execute(query)
        vf_infos = result.fetchall()
        if len(vf_infos) < 1:
            raise VFolderNotFound("Only project folders are directly sharable.")
        if len(vf_infos) > 1:
            raise InternalServerError(f"Multiple project folders found: {folder_name}")
        vf_info = vf_infos[0]
        allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
        await ensure_host_permission_allowed(
            conn,
            vf_info["host"],
            allowed_vfolder_types=allowed_vfolder_types,
            user_uuid=user_uuid,
            resource_policy=resource_policy,
            domain_name=domain_name,
            permission=VFolderHostPermission.SET_USER_PERM,
        )

        # Convert users' emails to uuids and check if user belong to the group of vfolder.
        j = users.join(agus, users.c.uuid == agus.c.user_id)
        query = (
            sa.select([users.c.uuid, users.c.email])
            .select_from(j)
            .where(
                (users.c.email.in_(params["emails"]))
                & (users.c.email != request["user"]["email"])
                & (agus.c.group_id == vf_info["group"]),
            )
        )
        result = await conn.execute(query)
        user_info = result.fetchall()
        users_to_share = [u["uuid"] for u in user_info]
        emails_to_share = [u["email"] for u in user_info]
        if len(user_info) < 1:
            raise ObjectNotFound(object_name="user")
        if len(user_info) < len(params["emails"]):
            users_not_in_vfolder_group = list(set(params["emails"]) - set(emails_to_share))
            raise ObjectNotFound(
                "Some user does not belong to folder's group: "
                ",".join(users_not_in_vfolder_group),
                object_name="user",
            )

        # Do not share to users who have already been shared the folder.
        query = (
            sa.select([vfolder_permissions.c.user])
            .select_from(vfolder_permissions)
            .where(
                (vfolder_permissions.c.user.in_(users_to_share))
                & (vfolder_permissions.c.vfolder == vf_info["id"]),
            )
        )
        result = await conn.execute(query)
        users_not_to_share = [u.user for u in result.fetchall()]
        users_to_share = list(set(users_to_share) - set(users_not_to_share))

        # Create vfolder_permission(s).
        for _user in users_to_share:
            query = sa.insert(
                vfolder_permissions,
                {
                    "permission": params["permission"],
                    "vfolder": vf_info["id"],
                    "user": _user,
                },
            )
            await conn.execute(query)
        # Update existing vfolder_permission(s).
        for _user in users_not_to_share:
            query = (
                sa.update(vfolder_permissions)
                .values(permission=params["permission"])
                .where(vfolder_permissions.c.vfolder == vf_info["id"])
                .where(vfolder_permissions.c.user == _user)
            )
            await conn.execute(query)

        return web.json_response({"shared_emails": emails_to_share}, status=201)


@admin_required
@server_status_required(ALL_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("emails"): t.List(t.String),
        }
    ),
)
async def unshare(request: web.Request, params: Any) -> web.Response:
    """
    Unshare a group folder from users.
    """
    await ensure_vfolder_status(
        request, VFolderAccessStatus.UPDATABLE, folder_name=request.match_info["name"]
    )
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    folder_name = request.match_info["name"]
    log.info(
        "VFOLDER.UNSHARE (email:{}, ak:{}, vf:{}, users:{})",
        request["user"]["email"],
        access_key,
        folder_name,
        ",".join(params["emails"]),
    )
    user_uuid = request["user"]["uuid"]
    domain_name = request["user"]["domain_name"]
    resource_policy = request["keypair"]["resource_policy"]
    async with root_ctx.db.begin() as conn:
        # Get the group-type virtual folder.
        query = (
            sa.select([vfolders.c.id, vfolders.c.host])
            .select_from(vfolders)
            .where(
                (vfolders.c.ownership_type == VFolderOwnershipType.GROUP)
                & (vfolders.c.name == folder_name),
            )
        )
        result = await conn.execute(query)
        vf_infos = result.fetchall()
        if len(vf_infos) < 1:
            raise VFolderNotFound("Only project folders are directly unsharable.")
        if len(vf_infos) > 1:
            raise InternalServerError(f"Multiple project folders found: {folder_name}")
        vf_info = vf_infos[0]
        allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
        await ensure_host_permission_allowed(
            conn,
            vf_info["host"],
            allowed_vfolder_types=allowed_vfolder_types,
            user_uuid=user_uuid,
            resource_policy=resource_policy,
            domain_name=domain_name,
            permission=VFolderHostPermission.SET_USER_PERM,
        )

        # Convert users' emails to uuids.
        query = (
            sa.select([users.c.uuid]).select_from(users).where(users.c.email.in_(params["emails"]))
        )
        result = await conn.execute(query)
        users_to_unshare = [u["uuid"] for u in result.fetchall()]
        if len(users_to_unshare) < 1:
            raise ObjectNotFound(object_name="user(s).")

        # Delete vfolder_permission(s).
        query = sa.delete(vfolder_permissions).where(
            (vfolder_permissions.c.vfolder == vf_info["id"])
            & (vfolder_permissions.c.user.in_(users_to_unshare)),
        )
        await conn.execute(query)
        return web.json_response({"unshared_emails": params["emails"]}, status=200)


@auth_required
@server_status_required(ALL_ALLOWED)
async def delete_by_name(request: web.Request) -> web.Response:
    await ensure_vfolder_status(
        request, VFolderAccessStatus.DELETABLE, folder_name=request.match_info["name"]
    )
    root_ctx: RootContext = request.app["_root.context"]
    app_ctx: PrivateContext = request.app["folders.context"]

    folder_name = request.match_info["name"]
    access_key = request["keypair"]["access_key"]
    domain_name = request["user"]["domain_name"]
    user_role = request["user"]["role"]
    user_uuid = request["user"]["uuid"]
    allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
    resource_policy = request["keypair"]["resource_policy"]

    log.info(
        "VFOLDER.DELETE (email:{}, ak:{}, vf:{})",
        request["user"]["email"],
        access_key,
        folder_name,
    )
    async with root_ctx.db.begin() as conn:
        entries = await query_accessible_vfolders(
            conn,
            user_uuid,
            user_role=user_role,
            domain_name=domain_name,
            allowed_vfolder_types=allowed_vfolder_types,
            extra_vf_conds=(vfolders.c.name == folder_name),
        )
        # for entry in entries:
        #     if entry['name'] == folder_name:
        #         # Folder owner OR user who have DELETE permission can delete folder.
        #         if (
        #             not entry['is_owner']
        #             and entry['permission'] != VFolderPermission.RW_DELETE
        #         ):
        #             raise InvalidAPIParameters(
        #                 'Cannot delete the vfolder '
        #                 'that is not owned by myself.')
        #         break
        # FIXME: For now, multiple entries on delete vfolder will raise an error. Will be fixed in 22.06
        if len(entries) > 1:
            log.error(
                "VFOLDER.DELETE(email:{}, folder name:{}, hosts:{}",
                request["user"]["email"],
                folder_name,
                [entry["host"] for entry in entries],
            )
            raise TooManyVFoldersFound(
                extra_msg="Multiple folders with the same name.",
                extra_data=None,
            )
        elif len(entries) == 0:
            raise InvalidAPIParameters("No such vfolder.")
        # query_accesible_vfolders returns list
        entry = entries[0]
        folder_host = entry["host"]
        await ensure_host_permission_allowed(
            conn,
            folder_host,
            allowed_vfolder_types=allowed_vfolder_types,
            user_uuid=user_uuid,
            resource_policy=resource_policy,
            domain_name=domain_name,
            permission=VFolderHostPermission.DELETE,
        )
        # Folder owner OR user who have DELETE permission can delete folder.
        if not entry["is_owner"] and entry["permission"] != VFolderPermission.RW_DELETE:
            raise InvalidAPIParameters("Cannot delete the vfolder that is not owned by myself.")

    await initiate_vfolder_removal(
        root_ctx.db,
        [VFolderDeletionInfo(entry["id"], folder_host)],
        root_ctx.storage_manager,
        app_ctx.storage_ptask_group,
    )
    return web.Response(status=204)


@auth_required
@server_status_required(ALL_ALLOWED)
@vfolder_permission_required(VFolderPermission.READ_ONLY)
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["shared_user_uuid", "sharedUserUuid"], default=None): t.String | t.Null,
        }
    ),
)
async def leave(request: web.Request, params: Any, row: VFolderRow) -> web.Response:
    """
    Leave a shared vfolder.

    Cannot leave a group vfolder or a vfolder that the requesting user owns.
    """
    await ensure_vfolder_status(
        request, VFolderAccessStatus.UPDATABLE, folder_name=request.match_info["name"]
    )
    if row["ownership_type"] == VFolderOwnershipType.GROUP:
        raise InvalidAPIParameters("Cannot leave a group vfolder.")

    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    user_role = request["user"]["role"]
    rqst_user_uuid = request["user"]["uuid"]
    shared_user_uuid = params["shared_user_uuid"]
    vfolder_id = row["id"]
    perm = row["permission"]

    if shared_user_uuid:
        # Allow only superadmin to leave the shared vfolder of others.
        if (rqst_user_uuid != shared_user_uuid) and (user_role != UserRole.SUPERADMIN):
            raise InsufficientPrivilege("Insufficient permission.")
        user_uuid = shared_user_uuid
    else:
        if row["is_owner"]:
            raise InvalidAPIParameters("Cannot leave a vfolder owned by the requesting user.")
        user_uuid = rqst_user_uuid

    log.info(
        "VFOLDER.LEAVE(email:{}, ak:{}, vfid:{}, uid:{}, perm:{})",
        request["user"]["email"],
        access_key,
        vfolder_id,
        user_uuid,
        perm,
    )
    async with root_ctx.db.begin() as conn:
        query = (
            sa.delete(vfolder_permissions)
            .where(vfolder_permissions.c.vfolder == vfolder_id)
            .where(vfolder_permissions.c.user == user_uuid)
        )
        await conn.execute(query)
    resp = {"msg": "left the shared vfolder"}
    return web.json_response(resp, status=200)


@auth_required
@server_status_required(ALL_ALLOWED)
@vfolder_permission_required(VFolderPermission.READ_ONLY)
@check_api_params(
    t.Dict(
        {
            t.Key("cloneable", default=False): t.Bool,
            t.Key("target_name"): tx.Slug(allow_dot=True),
            t.Key("target_host", default=None) >> "folder_host": t.String | t.Null,
            t.Key("usage_mode", default="general"): tx.Enum(VFolderUsageMode) | t.Null,
            t.Key("permission", default="rw"): tx.Enum(VFolderPermission) | t.Null,
        }
    ),
)
async def clone(request: web.Request, params: Any, row: VFolderRow) -> web.Response:
    await ensure_vfolder_status(
        request, VFolderAccessStatus.UPDATABLE, folder_name=request.match_info["name"]
    )
    resp: Dict[str, Any] = {}
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    user_role = request["user"]["role"]
    user_uuid = request["user"]["uuid"]
    resource_policy = request["keypair"]["resource_policy"]
    domain_name = request["user"]["domain_name"]
    log.info(
        "VFOLDER.CLONE (email:{}, ak:{}, vf:{}, vft:{}, vfh:{}, umod:{}, perm:{})",
        request["user"]["email"],
        access_key,
        row["name"],
        params["target_name"],
        params["folder_host"],
        params["usage_mode"].value,
        params["permission"].value,
    )
    source_folder_host = row["host"]
    source_folder_id = row["id"]
    target_folder_host = params["folder_host"]
    source_proxy_name, source_volume_name = root_ctx.storage_manager.split_host(source_folder_host)
    target_proxy_name, target_volume_name = root_ctx.storage_manager.split_host(target_folder_host)

    # check if the source vfolder is allowed to be cloned
    if not row["cloneable"]:
        raise GenericForbidden("The source vfolder is not permitted to be cloned.")

    if not target_folder_host:
        target_folder_host = await root_ctx.shared_config.etcd.get("volumes/default_host")
        if not target_folder_host:
            raise InvalidAPIParameters(
                "You must specify the vfolder host " "because the default host is not configured."
            )

    allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
    for vf_type in allowed_vfolder_types:
        if vf_type not in ("user", "group"):
            raise ServerMisconfiguredError(
                f"Invalid vfolder type(s): {str(allowed_vfolder_types)}."
                ' Only "user" or "group" is allowed.'
            )

    if not verify_vfolder_name(params["target_name"]):
        raise InvalidAPIParameters(f'{params["target_name"]} is reserved for internal operations.')

    if source_proxy_name != target_proxy_name:
        raise InvalidAPIParameters("proxy name of source and target vfolders must be equal.")

    async with root_ctx.db.begin() as conn:
        allowed_hosts = await filter_host_allowed_permission(
            conn,
            allowed_vfolder_types=allowed_vfolder_types,
            user_uuid=user_uuid,
            resource_policy=resource_policy,
            domain_name=domain_name,
        )
        if (
            target_folder_host not in allowed_hosts
            or VFolderHostPermission.CREATE not in allowed_hosts[target_folder_host]
        ):
            raise InvalidAPIParameters(
                f"`{VFolderHostPermission.CREATE}` Not allowed in vfolder host(`{target_folder_host}`)"
            )
        # TODO: handle legacy host lists assuming that volume names don't overlap?
        if target_folder_host not in allowed_hosts:
            raise InvalidAPIParameters("You are not allowed to use this vfolder host.")

        # Check resource policy's max_vfolder_count
        if resource_policy["max_vfolder_count"] > 0:
            query = sa.select([sa.func.count()]).where(vfolders.c.user == user_uuid)
            result = await conn.scalar(query)
            if result >= resource_policy["max_vfolder_count"]:
                raise InvalidAPIParameters("You cannot create more vfolders.")

        # Prevent creation of vfolder with duplicated name.
        extra_vf_conds = [vfolders.c.name == params["target_name"]]
        extra_vf_conds.append(vfolders.c.host == target_folder_host)
        entries = await query_accessible_vfolders(
            conn,
            user_uuid,
            user_role=user_role,
            domain_name=domain_name,
            allowed_vfolder_types=allowed_vfolder_types,
            extra_vf_conds=(sa.and_(*extra_vf_conds)),
        )
        if len(entries) > 0:
            raise VFolderAlreadyExists
        if params["target_name"].startswith("."):
            dotfiles, _ = await query_owned_dotfiles(conn, access_key)
            for dotfile in dotfiles:
                if params["target_name"] == dotfile["path"]:
                    raise InvalidAPIParameters("vFolder name conflicts with your dotfile.")

        if "user" not in allowed_vfolder_types:
            raise InvalidAPIParameters("user vfolder cannot be created in this host")

    task_id, target_folder_id = await initiate_vfolder_clone(
        root_ctx.db,
        VFolderCloneInfo(
            source_folder_id,
            source_folder_host,
            params["target_name"],
            target_folder_host,
            params["usage_mode"],
            params["permission"],
            request["user"]["email"],
            user_uuid,
            params["cloneable"],
        ),
        root_ctx.storage_manager,
        root_ctx.background_task_manager,
    )

    # Return the information about the destination vfolder.
    resp = {
        "id": target_folder_id.hex,
        "name": params["target_name"],
        "host": target_folder_host,
        "usage_mode": params["usage_mode"].value,
        "permission": params["permission"].value,
        "creator": request["user"]["email"],
        "ownership_type": "user",
        "user": str(user_uuid),
        "group": None,
        "cloneable": params["cloneable"],
        "bgtask_id": str(task_id),
    }
    return web.json_response(resp, status=201)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["vfolder_id", "vfolderId"], default=None): tx.UUID | t.Null,
        }
    ),
)
async def list_shared_vfolders(request: web.Request, params: Any) -> web.Response:
    """
    List shared vfolders.

    Not available for group vfolders.
    """
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    target_vfid = params["vfolder_id"]
    log.info(
        "VFOLDER.LIST_SHARED_VFOLDERS (email:{}, ak:{})",
        request["user"]["email"],
        access_key,
    )
    async with root_ctx.db.begin() as conn:
        j = vfolder_permissions.join(vfolders, vfolders.c.id == vfolder_permissions.c.vfolder).join(
            users, users.c.uuid == vfolder_permissions.c.user
        )
        query = sa.select(
            [
                vfolder_permissions,
                vfolders.c.id,
                vfolders.c.name,
                vfolders.c.group,
                vfolders.c.status,
                vfolders.c.user.label("vfolder_user"),
                users.c.email,
            ]
        ).select_from(j)
        if target_vfid is not None:
            query = query.where(vfolders.c.id == target_vfid)
        result = await conn.execute(query)
        shared_list = result.fetchall()
    shared_info = []
    for shared in shared_list:
        owner = shared.group if shared.group else shared.vfolder_user
        folder_type = "project" if shared.group else "user"
        shared_info.append(
            {
                "vfolder_id": str(shared.id),
                "vfolder_name": str(shared.name),
                "status": shared.status.value,
                "owner": str(owner),
                "type": folder_type,
                "shared_to": {
                    "uuid": str(shared.user),
                    "email": shared.email,
                },
                "perm": shared.permission.value,
            }
        )
    resp = {"shared": shared_info}
    return web.json_response(resp, status=200)


@auth_required
@server_status_required(ALL_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("vfolder"): tx.UUID,
            t.Key("user"): tx.UUID,
            tx.AliasedKey(["perm", "permission"]): VFolderPermissionValidator | t.Null,
        }
    ),
)
async def update_shared_vfolder(request: web.Request, params: Any) -> web.Response:
    """
    Update permission for shared vfolders.

    If params['perm'] is None, remove user's permission for the vfolder.
    """
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    vfolder_id = params["vfolder"]
    user_uuid = params["user"]
    perm = params["perm"]
    log.info(
        "VFOLDER.UPDATE_SHARED_VFOLDER(email:{}, ak:{}, vfid:{}, uid:{}, perm:{})",
        request["user"]["email"],
        access_key,
        vfolder_id,
        user_uuid,
        perm,
    )
    async with root_ctx.db.begin() as conn:
        if perm is not None:
            query = (
                sa.update(vfolder_permissions)
                .values(permission=perm)
                .where(vfolder_permissions.c.vfolder == vfolder_id)
                .where(vfolder_permissions.c.user == user_uuid)
            )
        else:
            query = (
                sa.delete(vfolder_permissions)
                .where(vfolder_permissions.c.vfolder == vfolder_id)
                .where(vfolder_permissions.c.user == user_uuid)
            )
        await conn.execute(query)
    resp = {"msg": "shared vfolder permission updated"}
    return web.json_response(resp, status=200)


@superadmin_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("fstab_path", default=None): t.String | t.Null,
            t.Key("agent_id", default=None): t.String | t.Null,
        }
    ),
)
async def get_fstab_contents(request: web.Request, params: Any) -> web.Response:
    """
    Return the contents of `/etc/fstab` file.
    """
    access_key = request["keypair"]["access_key"]
    log.info(
        "VFOLDER.GET_FSTAB_CONTENTS(email:{}, ak:{}, ag:{})",
        request["user"]["email"],
        access_key,
        params["agent_id"],
    )
    if params["fstab_path"] is None:
        params["fstab_path"] = "/etc/fstab"
    if params["agent_id"] is not None:
        # Return specific agent's fstab.
        watcher_info = await get_watcher_info(request, params["agent_id"])
        try:
            client_timeout = aiohttp.ClientTimeout(total=10.0)
            async with aiohttp.ClientSession(timeout=client_timeout) as sess:
                headers = {"X-BackendAI-Watcher-Token": watcher_info["token"]}
                url = watcher_info["addr"] / "fstab"
                async with sess.get(url, headers=headers, params=params) as watcher_resp:
                    if watcher_resp.status == 200:
                        content = await watcher_resp.text()
                        resp = {
                            "content": content,
                            "node": "agent",
                            "node_id": params["agent_id"],
                        }
                        return web.json_response(resp)
                    else:
                        message = await watcher_resp.text()
                        raise BackendAgentError(
                            "FAILURE", f"({watcher_resp.status}: {watcher_resp.reason}) {message}"
                        )
        except asyncio.CancelledError:
            raise
        except asyncio.TimeoutError:
            log.error(
                "VFOLDER.GET_FSTAB_CONTENTS(u:{}): timeout from watcher (agent:{})",
                access_key,
                params["agent_id"],
            )
            raise BackendAgentError("TIMEOUT", "Could not fetch fstab data from agent")
        except Exception:
            log.exception(
                "VFOLDER.GET_FSTAB_CONTENTS(u:{}): "
                "unexpected error while reading from watcher (agent:{})",
                access_key,
                params["agent_id"],
            )
            raise InternalServerError
    else:
        resp = {
            "content": "# Since Backend.AI 20.09, reading the manager fstab is no longer supported.",
            "node": "manager",
            "node_id": "manager",
        }
        return web.json_response(resp)


@superadmin_required
@server_status_required(READ_ALLOWED)
async def list_mounts(request: web.Request) -> web.Response:
    """
    List all mounted vfolder hosts in vfroot.

    All mounted hosts from connected (ALIVE) agents are also gathered.
    Generally, agents should be configured to have same hosts structure,
    but newly introduced one may not.
    """
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    log.info("VFOLDER.LIST_MOUNTS(ak:{})", access_key)
    mount_prefix = await root_ctx.shared_config.get_raw("volumes/_mount")
    if mount_prefix is None:
        mount_prefix = "/mnt"

    # NOTE: Changed in 20.09: the manager instances no longer have mountpoints.
    all_volumes = [*await root_ctx.storage_manager.get_all_volumes()]
    all_mounts = [volume_data["path"] for proxy_name, volume_data in all_volumes]
    all_vfolder_hosts = [
        f"{proxy_name}:{volume_data['name']}" for proxy_name, volume_data in all_volumes
    ]
    resp: MutableMapping[str, Any] = {
        "manager": {
            "success": True,
            "mounts": all_mounts,
            "message": "(legacy)",
        },
        "storage-proxy": {
            "success": True,
            "mounts": [*zip(all_vfolder_hosts, all_mounts)],
            "message": "",
        },
        "agents": {},
    }

    # Scan mounted vfolder hosts for connected agents.
    async def _fetch_mounts(
        sema: asyncio.Semaphore,
        sess: aiohttp.ClientSession,
        agent_id: str,
    ) -> Tuple[str, Mapping]:
        async with sema:
            watcher_info = await get_watcher_info(request, agent_id)
            headers = {"X-BackendAI-Watcher-Token": watcher_info["token"]}
            url = watcher_info["addr"] / "mounts"
            try:
                async with sess.get(url, headers=headers) as watcher_resp:
                    if watcher_resp.status == 200:
                        data = {
                            "success": True,
                            "mounts": await watcher_resp.json(),
                            "message": "",
                        }
                    else:
                        data = {
                            "success": False,
                            "mounts": [],
                            "message": await watcher_resp.text(),
                        }
                    return (agent_id, data)
            except asyncio.CancelledError:
                raise
            except asyncio.TimeoutError:
                log.error(
                    "VFOLDER.LIST_MOUNTS(u:{}): timeout from watcher (agent:{})",
                    access_key,
                    agent_id,
                )
                raise
            except Exception:
                log.exception(
                    "VFOLDER.LIST_MOUNTS(u:{}): "
                    "unexpected error while reading from watcher (agent:{})",
                    access_key,
                    agent_id,
                )
                raise

    async with root_ctx.db.begin() as conn:
        query = (
            sa.select([agents.c.id]).select_from(agents).where(agents.c.status == AgentStatus.ALIVE)
        )
        result = await conn.execute(query)
        rows = result.fetchall()

    client_timeout = aiohttp.ClientTimeout(total=10.0)
    async with aiohttp.ClientSession(timeout=client_timeout) as sess:
        sema = asyncio.Semaphore(8)
        mounts = await asyncio.gather(
            *[_fetch_mounts(sema, sess, row.id) for row in rows], return_exceptions=True
        )
        for mount in mounts:
            if isinstance(mount, Exception):
                # exceptions are already logged.
                continue
            resp["agents"][mount[0]] = mount[1]

    return web.json_response(resp, status=200)


@superadmin_required
@server_status_required(ALL_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("fs_location"): t.String,
            t.Key("name"): t.String,
            t.Key("fs_type", default="nfs"): t.String,
            t.Key("options", default=None): t.String | t.Null,
            t.Key("scaling_group", default=None): t.String | t.Null,
            t.Key("fstab_path", default=None): t.String | t.Null,
            t.Key("edit_fstab", default=False): t.ToBool,
        }
    ),
)
async def mount_host(request: web.Request, params: Any) -> web.Response:
    """
    Mount device into vfolder host.

    Mount a device (eg: nfs) located at `fs_location` into `<vfroot>/name` in the
    host machines (manager and all agents). `fs_type` can be specified by requester,
    which fallbaks to 'nfs'.

    If `scaling_group` is specified, try to mount for agents in the scaling group.
    """
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    log_fmt = "VFOLDER.MOUNT_HOST(ak:{}, name:{}, fs:{}, sg:{})"
    log_args = (access_key, params["name"], params["fs_location"], params["scaling_group"])
    log.info(log_fmt, *log_args)
    mount_prefix = await root_ctx.shared_config.get_raw("volumes/_mount")
    if mount_prefix is None:
        mount_prefix = "/mnt"

    # NOTE: Changed in 20.09: the manager instances no longer have mountpoints.
    resp: MutableMapping[str, Any] = {
        "manager": {
            "success": True,
            "message": "Managers do not have mountpoints since v20.09.",
        },
        "agents": {},
    }

    # Mount on running agents.
    async with root_ctx.db.begin() as conn:
        query = (
            sa.select([agents.c.id]).select_from(agents).where(agents.c.status == AgentStatus.ALIVE)
        )
        if params["scaling_group"] is not None:
            query = query.where(agents.c.scaling == params["scaling_group"])
        result = await conn.execute(query)
        rows = result.fetchall()

    async def _mount(
        sema: asyncio.Semaphore,
        sess: aiohttp.ClientSession,
        agent_id: str,
    ) -> Tuple[str, Mapping]:
        async with sema:
            watcher_info = await get_watcher_info(request, agent_id)
            try:
                headers = {"X-BackendAI-Watcher-Token": watcher_info["token"]}
                url = watcher_info["addr"] / "mounts"
                async with sess.post(url, json=params, headers=headers) as resp:
                    if resp.status == 200:
                        data = {
                            "success": True,
                            "message": await resp.text(),
                        }
                    else:
                        data = {
                            "success": False,
                            "message": await resp.text(),
                        }
                    return (agent_id, data)
            except asyncio.CancelledError:
                raise
            except asyncio.TimeoutError:
                log.error(
                    log_fmt + ": timeout from watcher (ag:{})",
                    *log_args,
                    agent_id,
                )
                raise
            except Exception:
                log.exception(
                    log_fmt + ": unexpected error while reading from watcher (ag:{})",
                    *log_args,
                    agent_id,
                )
                raise

    client_timeout = aiohttp.ClientTimeout(total=10)
    async with aiohttp.ClientSession(timeout=client_timeout) as sess:
        sema = asyncio.Semaphore(8)
        results = await asyncio.gather(
            *[_mount(sema, sess, row.id) for row in rows], return_exceptions=True
        )
        for result in results:
            if isinstance(result, Exception):
                # exceptions are already logged.
                continue
            resp["agents"][result[0]] = result[1]

    return web.json_response(resp, status=200)


@superadmin_required
@server_status_required(ALL_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("name"): t.String,
            t.Key("scaling_group", default=None): t.String | t.Null,
            t.Key("fstab_path", default=None): t.String | t.Null,
            t.Key("edit_fstab", default=False): t.ToBool,
        }
    ),
)
async def umount_host(request: web.Request, params: Any) -> web.Response:
    """
    Unmount device from vfolder host.

    Unmount a device (eg: nfs) located at `<vfroot>/name` from the host machines
    (manager and all agents).

    If `scaling_group` is specified, try to unmount for agents in the scaling group.
    """
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    log_fmt = "VFOLDER.UMOUNT_HOST(ak:{}, name:{}, sg:{})"
    log_args = (access_key, params["name"], params["scaling_group"])
    log.info(log_fmt, *log_args)
    mount_prefix = await root_ctx.shared_config.get_raw("volumes/_mount")
    if mount_prefix is None:
        mount_prefix = "/mnt"
    mountpoint = Path(mount_prefix) / params["name"]
    assert Path(mount_prefix) != mountpoint

    async with root_ctx.db.begin() as conn, conn.begin():
        # Prevent unmount if target host is mounted to running kernels.
        query = (
            sa.select([kernels.c.mounts])
            .select_from(kernels)
            .where(kernels.c.status != KernelStatus.TERMINATED)
        )
        result = await conn.execute(query)
        _kernels = result.fetchall()
        _mounted = set()
        for kern in _kernels:
            if kern.mounts:
                _mounted.update([m[1] for m in kern.mounts])
        if params["name"] in _mounted:
            return web.json_response(
                {
                    "title": "Target host is used in sessions",
                    "message": "Target host is used in sessions",
                },
                status=409,
            )

        query = (
            sa.select([agents.c.id]).select_from(agents).where(agents.c.status == AgentStatus.ALIVE)
        )
        if params["scaling_group"] is not None:
            query = query.where(agents.c.scaling == params["scaling_group"])
        result = await conn.execute(query)
        _agents = result.fetchall()

    # Unmount from manager.
    # NOTE: Changed in 20.09: the manager instances no longer have mountpoints.
    resp: MutableMapping[str, Any] = {
        "manager": {
            "success": True,
            "message": "Managers do not have mountpoints since v20.09.",
        },
        "agents": {},
    }

    # Unmount from running agents.
    async def _umount(
        sema: asyncio.Semaphore,
        sess: aiohttp.ClientSession,
        agent_id: str,
    ) -> Tuple[str, Mapping]:
        async with sema:
            watcher_info = await get_watcher_info(request, agent_id)
            try:
                headers = {"X-BackendAI-Watcher-Token": watcher_info["token"]}
                url = watcher_info["addr"] / "mounts"
                async with sess.delete(url, json=params, headers=headers) as resp:
                    if resp.status == 200:
                        data = {
                            "success": True,
                            "message": await resp.text(),
                        }
                    else:
                        data = {
                            "success": False,
                            "message": await resp.text(),
                        }
                    return (agent_id, data)
            except asyncio.CancelledError:
                raise
            except asyncio.TimeoutError:
                log.error(
                    log_fmt + ": timeout from watcher (agent:{})",
                    *log_args,
                    agent_id,
                )
                raise
            except Exception:
                log.exception(
                    log_fmt + ": unexpected error while reading from watcher (agent:{})",
                    *log_args,
                    agent_id,
                )
                raise

    client_timeout = aiohttp.ClientTimeout(total=10.0)
    async with aiohttp.ClientSession(timeout=client_timeout) as sess:
        sema = asyncio.Semaphore(8)
        results = await asyncio.gather(
            *[_umount(sema, sess, _agent.id) for _agent in _agents], return_exceptions=True
        )
        for result in results:
            if isinstance(result, Exception):
                # exceptions are already logged.
                continue
            resp["agents"][result[0]] = result[1]

    return web.json_response(resp, status=200)


async def storage_task_exception_handler(
    exc_type: Type[Exception],
    exc_obj: Exception,
    tb: TracebackType,
):
    log.exception("Error while removing vFolder", exc_info=exc_obj)


@attrs.define(slots=True, auto_attribs=True, init=False)
class PrivateContext:
    database_ptask_group: aiotools.PersistentTaskGroup
    storage_ptask_group: aiotools.PersistentTaskGroup


async def init(app: web.Application) -> None:
    app_ctx: PrivateContext = app["folders.context"]
    app_ctx.database_ptask_group = aiotools.PersistentTaskGroup()
    app_ctx.storage_ptask_group = aiotools.PersistentTaskGroup(
        exception_handler=storage_task_exception_handler
    )


async def shutdown(app: web.Application) -> None:
    app_ctx: PrivateContext = app["folders.context"]
    await app_ctx.database_ptask_group.shutdown()
    await app_ctx.storage_ptask_group.shutdown()


def create_app(default_cors_options):
    app = web.Application()
    app["prefix"] = "folders"
    app["api_versions"] = (2, 3, 4)
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app["folders.context"] = PrivateContext()
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    add_route = app.router.add_route
    root_resource = cors.add(app.router.add_resource(r""))
    cors.add(root_resource.add_route("POST", create))
    cors.add(root_resource.add_route("GET", list_folders))
    cors.add(root_resource.add_route("DELETE", delete_by_id))
    vfolder_resource = cors.add(app.router.add_resource(r"/{name}"))
    cors.add(vfolder_resource.add_route("GET", get_info))
    cors.add(vfolder_resource.add_route("DELETE", delete_by_name))
    cors.add(add_route("GET", r"/_/hosts", list_hosts))
    cors.add(add_route("GET", r"/_/all-hosts", list_all_hosts))
    cors.add(add_route("GET", r"/_/allowed-types", list_allowed_types))
    cors.add(add_route("GET", r"/_/all_hosts", list_all_hosts))  # legacy underbar
    cors.add(add_route("GET", r"/_/allowed_types", list_allowed_types))  # legacy underbar
    cors.add(add_route("GET", r"/_/perf-metric", get_volume_perf_metric))
    cors.add(add_route("POST", r"/{name}/rename", rename_vfolder))
    cors.add(add_route("POST", r"/{name}/update-options", update_vfolder_options))
    cors.add(add_route("POST", r"/{name}/mkdir", mkdir))
    cors.add(add_route("POST", r"/{name}/request-upload", create_upload_session))
    cors.add(add_route("POST", r"/{name}/request-download", create_download_session))
    cors.add(add_route("POST", r"/{name}/move-file", move_file))
    cors.add(add_route("POST", r"/{name}/rename-file", rename_file))
    cors.add(add_route("DELETE", r"/{name}/delete-files", delete_files))
    cors.add(add_route("POST", r"/{name}/rename_file", rename_file))  # legacy underbar
    cors.add(add_route("DELETE", r"/{name}/delete_files", delete_files))  # legacy underbar
    cors.add(add_route("GET", r"/{name}/files", list_files))
    cors.add(add_route("POST", r"/{name}/invite", invite))
    cors.add(add_route("POST", r"/{name}/leave", leave))
    cors.add(add_route("POST", r"/{name}/share", share))
    cors.add(add_route("DELETE", r"/{name}/unshare", unshare))
    cors.add(add_route("POST", r"/{name}/clone", clone))
    cors.add(add_route("GET", r"/invitations/list-sent", list_sent_invitations))
    cors.add(add_route("GET", r"/invitations/list_sent", list_sent_invitations))  # legacy underbar
    cors.add(add_route("POST", r"/invitations/update/{inv_id}", update_invitation))
    cors.add(add_route("GET", r"/invitations/list", invitations))
    cors.add(add_route("POST", r"/invitations/accept", accept_invitation))
    cors.add(add_route("DELETE", r"/invitations/delete", delete_invitation))
    cors.add(add_route("GET", r"/_/shared", list_shared_vfolders))
    cors.add(add_route("POST", r"/_/shared", update_shared_vfolder))
    cors.add(add_route("GET", r"/_/fstab", get_fstab_contents))
    cors.add(add_route("GET", r"/_/mounts", list_mounts))
    cors.add(add_route("POST", r"/_/mounts", mount_host))
    cors.add(add_route("DELETE", r"/_/mounts", umount_host))
    cors.add(add_route("GET", r"/_/quota", get_quota))
    cors.add(add_route("POST", r"/_/quota", update_quota))
    cors.add(add_route("GET", r"/_/usage", get_usage))
    cors.add(add_route("GET", r"/_/used-bytes", get_used_bytes))
    return app, []



import uuid
from typing import TYPE_CHECKING, Sequence

import graphene
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql as psql
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import relationship
from sqlalchemy.orm.exc import NoResultFound

from ..api.exceptions import RoutingNotFound
from .base import GUID, Base, IDColumn, Item, PaginatedList
from .utils import ExtendedAsyncSAEngine, execute_with_retry

if TYPE_CHECKING:
    # from .gql import GraphQueryContext
    pass


__all__ = ("RoutingRow", "Routing", "RoutingList")


class RoutingRow(Base):
    __tablename__ = "routings"
    __table_args__ = (
        sa.UniqueConstraint("endpoint", "session", name="uq_routings_endpoint_session"),
    )

    id = IDColumn("id")
    endpoint = sa.Column(
        "endpoint", GUID, sa.ForeignKey("endpoints.id", ondelete="CASCADE"), nullable=False
    )
    session = sa.Column(
        "session", GUID, sa.ForeignKey("sessions.id", ondelete="RESTRICT"), nullable=False
    )

    traffic_ratio = sa.Column("traffic_ratio", sa.Float(), nullable=False)

    endpoint_row = relationship("EndpointRow", back_populates="routings")

    @classmethod
    async def get(cls, session: AsyncSession, routing_id: uuid.UUID) -> "RoutingRow":
        """
        :raises: sqlalchemy.orm.exc.NoResultFound
        """
        query = sa.select(RoutingRow).filter(RoutingRow.id == routing_id)
        result = await session.execute(query)
        try:
            return result.one()
        except NoResultFound:
            raise

    @classmethod
    async def create(
        cls,
        engine: ExtendedAsyncSAEngine,
        endpoint_id: uuid.UUID,
        session_id: uuid.UUID,
        traffic_ratio: float = 100.0,
    ) -> uuid.UUID:
        # https://docs.sqlalchemy.org/en/14/dialects/postgresql.html#sqlalchemy.dialects.postgresql.Insert.on_conflict_do_nothing

        async def _create_routing() -> uuid.UUID:
            async with engine.begin_session() as db_sess:
                routing_id = uuid.uuid4()
                query = (
                    psql.insert(RoutingRow)
                    .values(
                        id=routing_id,
                        endpoint=endpoint_id,
                        session=session_id,
                        traffic_ratio=traffic_ratio,
                    )
                    .on_conflict_do_nothing(
                        index_elements=[RoutingRow.endpoint, RoutingRow.session]
                    )
                )
                await db_sess.execute(query)
                return routing_id

        return await execute_with_retry(_create_routing)


class Routing(graphene.ObjectType):
    class Meta:
        interfaces = (Item,)

    routing_id = graphene.UUID()
    endpoint = graphene.String()
    session = graphene.UUID()
    traffic_ratio = graphene.Float()

    @classmethod
    async def from_row(
        cls,
        ctx,  # ctx: GraphQueryContext,
        row: RoutingRow,
    ) -> "Routing":
        return cls(
            routing_id=row.id,
            endpoint=row.endpoint_row.url,
            session=row.session,
            traffic_ratio=row.traffic_ratio,
        )

    @classmethod
    async def load_count(
        cls,
        ctx,  # ctx: GraphQueryContext,
        *,
        endpoint_id: uuid.UUID | None = None,
    ) -> int:
        query = sa.select([sa.func.count()]).select_from(RoutingRow)
        if endpoint_id is not None:
            query = query.where(RoutingRow.endpoint == endpoint_id)
        async with ctx.db.begin_readonly() as conn:
            result = await conn.execute(query)
            return result.scalar()

    @classmethod
    async def load_slice(
        cls,
        ctx,  # ctx: GraphQueryContext,
        limit: int,
        offset: int,
        *,
        endpoint_id: uuid.UUID | None = None,
        filter: str | None = None,
        order: str | None = None,
    ) -> Sequence["Routing"]:
        query = sa.select(RoutingRow).limit(limit).offset(offset)
        if endpoint_id is not None:
            query = query.where(RoutingRow.endpoint == endpoint_id)
        """
        if filter is not None:
            parser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = parser.append_filter(query, filter)
        if order is not None:
            parser = QueryOrderParser(cls._queryorder_colmap)
            query = parser.append_ordering(query, order)
        """
        async with ctx.db.begin_readonly_session() as session:
            return [await cls.from_row(ctx, row) async for row in (await session.stream(query))]

    @classmethod
    async def load_all(
        cls, ctx, *, endpoint_id: uuid.UUID | None = None  # ctx: GraphQueryContext,
    ) -> Sequence["Routing"]:
        async with ctx.db.begin_readonly_session() as session:
            rows = await RoutingRow.list(session, endpoint_id=endpoint_id)
        return [await Routing.from_row(ctx, row) for row in rows]

    @classmethod
    async def load_item(
        cls,
        ctx,  # ctx: GraphQueryContext,
        *,
        routing_id: uuid.UUID,
    ) -> "Routing":
        try:
            async with ctx.db.begin_readonly_session() as session:
                row = await RoutingRow.get(session, routing_id=routing_id)
        except NoResultFound:
            raise RoutingNotFound
        return await Routing.from_row(ctx, row)


class RoutingList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(Routing, required=True)



import asyncio
import logging
import secrets
import uuid
from datetime import datetime
from typing import TYPE_CHECKING, Any, Iterable, MutableMapping, Tuple, Union

import aiohttp_cors
import aiotools
import attrs
import sqlalchemy as sa
import trafaret as t
from aiohttp import web
from dateutil.parser import isoparse
from dateutil.tz import tzutc

from ai.backend.common import validators as tx
from ai.backend.common.docker import ImageRef
from ai.backend.common.exception import AliasResolutionFailed, UnknownImageReference
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import ClusterMode, SessionTypes
from ai.backend.common.utils import str_to_timedelta

from ..defs import DEFAULT_IMAGE_ARCH, DEFAULT_ROLE
from ..models import (
    ImageRow,
    UserRole,
    domains,
    query_bootstrap_script,
    verify_vfolder_name,
    vfolders,
)
from ..models.endpoint import EndpointRow
from ..models.routing import RoutingRow
from ..models.session import SessionRow, SessionStatus
from ..models.utils import execute_with_retry
from ..types import UserScope
from .auth import auth_required
from .exceptions import (
    BackendError,
    ImageNotFound,
    InsufficientPrivilege,
    InternalServerError,
    InvalidAPIParameters,
    UnknownImageReferenceError,
)
from .manager import ALL_ALLOWED, READ_ALLOWED, server_status_required
from .session import query_userinfo
from .types import CORSOptions, WebMiddleware
from .utils import check_api_params, get_access_key_scopes, undefined

if TYPE_CHECKING:
    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__name__))


class UndefChecker(t.Trafaret):
    def check_and_return(self, value: Any) -> object:
        if value == undefined:
            return value
        else:
            self._failure("Invalid Undef format", value=value)
            return None


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["project_id", "projectId"]): tx.UUID,
        }
    ),
)
async def list_serve(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    project_id = params["project_id"]

    log.info("SERVE.LIST (email:{}, ak:{})", request["user"]["email"], access_key)

    async with root_ctx.db.begin_readonly_session() as db_sess:
        j = sa.join(SessionRow, RoutingRow, SessionRow.id == RoutingRow.session).join(
            EndpointRow, RoutingRow.endpoint == EndpointRow.id
        )
        query = (
            sa.select(
                SessionRow.id,
                SessionRow.name,
                SessionRow.status,
            )
            .select_from(j)
            .where(SessionRow.group_id == project_id)
        )
        result = await db_sess.execute(query)
        rows = result.scalars().all()

    return web.json_response(rows, status=200)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["endpoint_id", "endpointId"]): tx.UUID,
        }
    ),
)
async def get_info(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]

    log.info("SERVE.GET_INFO (email:{}, ak:{})", request["user"]["email"], access_key)

    async with root_ctx.db.begin_readonly_session() as db_sess:
        j = sa.join(SessionRow, RoutingRow, SessionRow.id == RoutingRow.session).join(
            EndpointRow, RoutingRow.endpoint == EndpointRow.id
        )
        query = (
            sa.select(
                SessionRow.id,
                SessionRow.name,
                SessionRow.status,
                EndpointRow.image,
                EndpointRow.model,
                EndpointRow.url,
            )
            .select_from(j)
            .where(RoutingRow.endpoint == params["endpoint_id"])
        )
        result = await db_sess.execute(query)
        rows = result.scalars().all()

    return web.json_response(rows, status=200)


@auth_required
@server_status_required(ALL_ALLOWED)
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["endpoint_id", "endpointId"], default=None): tx.UUID | t.Null,
            tx.AliasedKey(["service_name", "serviceName"]): t.String,
            tx.AliasedKey(["model_id", "modelId"]): tx.UUID,
            tx.AliasedKey(["model_version", "modelVersion"]): t.String,
            tx.AliasedKey(["image_ref", "imageRef"]): t.String,
            tx.AliasedKey(["arch", "architecture"], default=DEFAULT_IMAGE_ARCH)
            >> "architecture": t.String,
            tx.AliasedKey(
                ["group", "groupName", "group_name", "project", "project_name", "projectName"],
                default="default",
            ): t.String,
            tx.AliasedKey(["domain", "domainName", "domain_name"], default="default"): t.String,
            tx.AliasedKey(["resource_opts", "resourceOpts"], default=dict): t.Mapping(
                t.String, t.Any
            ),
            tx.AliasedKey(["cluster_size", "clusterSize"], default=1): t.ToInt[1:],
            tx.AliasedKey(["cluster_mode", "clusterMode"], default="single-node"): tx.Enum(
                ClusterMode
            ),
            t.Key("tag", default=None): t.Null | t.String,
            t.Key("config", default=dict): t.Mapping(t.String, t.Any),
        }
    ),
)
async def create(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    app_ctx: PrivateContext = request.app["services.context"]
    access_key = request["keypair"]["access_key"]
    domain_name = request["user"]["domain_name"]
    user_role = request["user"]["role"]
    model_id = params["model_id"]

    params["owner_access_key"] = access_key
    async with root_ctx.db.begin_readonly() as db_conn:
        owner_uuid, group_id, resource_policy = await query_userinfo(request, params, db_conn)

    log.info("SERVE.CREATE (email:{}, ak:{})", request["user"]["email"], access_key)
    resp: MutableMapping[str, Any] = {}

    # Resolve the image reference.
    try:
        async with root_ctx.db.begin_readonly_session() as session:
            image_row = await ImageRow.resolve(
                session,
                [
                    ImageRef(params["image_ref"], ["*"], params["architecture"]),
                    params["image_ref"],
                ],
            )
            img_id = image_row.id
        requested_image_ref = image_row.image_ref
        model_mount_path = image_row.labels.get("ai.backend.model-path")
        async with root_ctx.db.begin_readonly() as conn:
            query = (
                sa.select([domains.c.allowed_docker_registries])
                .select_from(domains)
                .where(domains.c.name == params["domain"])
            )
            allowed_registries = await conn.scalar(query)
            if requested_image_ref.registry not in allowed_registries:
                raise AliasResolutionFailed
    except AliasResolutionFailed:
        raise ImageNotFound("unknown alias or disallowed registry")

    # Check work directory and reserved name directory.
    mount_map = params["config"].get("mount_map", {})
    if mount_map is not None:
        original_folders = mount_map.keys()
        alias_folders = mount_map.values()
        if len(alias_folders) != len(set(alias_folders)):
            raise InvalidAPIParameters("Duplicate alias folder name exists.")

        alias_name: str
        for alias_name in alias_folders:
            if alias_name is None:
                continue
            if alias_name.startswith("/home/work/"):
                alias_name = alias_name.replace("/home/work/", "")
            if alias_name == "":
                raise InvalidAPIParameters("Alias name cannot be empty.")
            if alias_name == model_mount_path:
                raise InvalidAPIParameters(
                    f"Alias name cannot be the same with model path: `{model_mount_path}`"
                )
            if not verify_vfolder_name(alias_name):
                raise InvalidAPIParameters(str(alias_name) + " is reserved for internal path.")
            if alias_name in original_folders:
                raise InvalidAPIParameters(
                    f"Alias name cannot be set to an existing folder name: {alias_name}"
                )

    # Append model mount path
    mounts: list[str] = params["config"].get("mounts", [])
    async with root_ctx.db.begin_readonly() as db_conn:
        query = sa.select([vfolders.c.name]).where(vfolders.c.id == model_id)
        result = await db_conn.execute(query)
        vfolder_name = result.first()["name"]
    model_path = f"{vfolder_name}/versions/{params['model_version']}"
    mounts = [*mounts, model_path]
    mount_map = {
        **mount_map,
        model_path: model_mount_path,
    }
    params["config"]["mounts"] = mounts
    params["config"]["mount_map"] = mount_map
    for kern_config in params["config"]["kernel_configs"]:
        kern_config["mounts"] = mounts
        kern_config["mount_map"] = mount_map

    # Create endpoint
    async def _create_endpoint() -> uuid.UUID:
        async with root_ctx.db.begin_session() as db_sess:
            endpoint_id = uuid.uuid4()
            endpoint = EndpointRow(
                id=endpoint_id,
                image=img_id,
                model=model_id,
                domain=domain_name,
                project=group_id,
                resource_group=params["config"]["scaling_group"],
            )
            db_sess.add(endpoint)
            return endpoint_id

    if params["endpoint_id"] is None:
        endpoint_id = await execute_with_retry(_create_endpoint)
    else:
        endpoint_id = params["endpoint_id"]

    # # Check existing (owner_access_key, session_name) instance
    # try:
    #     # NOTE: We can reuse the session IDs of TERMINATED sessions only.
    #     # NOTE: Reusing a session in the PENDING status returns an empty value in service_ports.
    #     async with root_ctx.db.begin_readonly_session() as db_sess:
    #         sess = await SessionRow.get_session_with_main_kernel(
    #             params["session_name"],
    #             access_key,
    #             db_session=db_sess,
    #         )
    #     running_image_ref = ImageRef(
    #         sess.main_kernel.image, [sess.main_kernel.registry], sess.main_kernel.architecture
    #     )
    #     if running_image_ref != requested_image_ref:
    #         # The image must be same if get_or_create() called multiple times
    #         # against an existing (non-terminated) session
    #         raise SessionAlreadyExists(extra_data={"existingSessionId": str(sess.id)})
    #     if not params["reuse"]:
    #         # Respond as error since the client did not request to reuse,
    #         # but provide the overlapping session ID for later use.
    #         raise SessionAlreadyExists(extra_data={"existingSessionId": str(sess.id)})
    #     # Respond as success with the reused session's information.
    #     return web.json_response(
    #         {
    #             "sessionId": str(sess.id),
    #             "sessionName": str(sess.name),
    #             "status": sess.status.name,
    #             "service_ports": sess.main_kernel.service_ports,
    #             "created": False,
    #         },
    #         status=200,
    #     )
    # except SessionNotFound:
    #     # It's time to create a new session.
    #     pass

    # if params["session_type"] == SessionTypes.BATCH and not params["startup_command"]:
    #     raise InvalidAPIParameters("Batch sessions must have a non-empty startup command.")
    # if params["session_type"] != SessionTypes.BATCH and params["starts_at"]:
    #     raise InvalidAPIParameters("Parameter starts_at should be used only for batch sessions")

    starts_at: Union[datetime, None] = None
    if params["starts_at"]:
        try:
            starts_at = isoparse(params["starts_at"])
        except ValueError:
            _td = str_to_timedelta(params["starts_at"])
            starts_at = datetime.now(tzutc()) + _td

    # if params["cluster_size"] > 1:
    #     log.debug(" -> cluster_mode:{} (replicate)", params["cluster_mode"])

    # if params["dependencies"] is None:
    #     params["dependencies"] = []

    session_creation_id = secrets.token_urlsafe(16)

    async with root_ctx.db.begin_readonly() as conn:
        # Use keypair bootstrap_script if it is not delivered as a parameter
        if not params["bootstrap_script"]:
            script, _ = await query_bootstrap_script(conn, access_key)
            params["bootstrap_script"] = script

    try:
        session_id = await asyncio.shield(
            app_ctx.database_ptask_group.create_task(
                root_ctx.registry.enqueue_session(
                    session_creation_id,
                    params["service_name"],
                    access_key,
                    {
                        "creation_config": params["config"],
                        "kernel_configs": [
                            {
                                "image_ref": requested_image_ref,
                                "cluster_role": DEFAULT_ROLE,
                                "cluster_idx": 1,
                                "local_rank": 0,
                                "cluster_hostname": f"{DEFAULT_ROLE}1",
                                "creation_config": params["config"],
                                "bootstrap_script": params["bootstrap_script"],
                                "startup_command": params["startup_command"],
                            }
                        ],
                    },
                    params["config"]["scaling_group"],
                    SessionTypes.INFERENCE,
                    resource_policy,
                    user_scope=UserScope(
                        domain_name=domain_name,
                        group_id=group_id,
                        user_uuid=owner_uuid,
                        user_role=user_role,
                    ),
                    cluster_mode=params["cluster_mode"],
                    cluster_size=params["cluster_size"],
                    session_tag=params["tag"],
                    starts_at=starts_at,
                    agent_list=params["config"]["agent_list"],
                    dependency_sessions=params["dependencies"],
                )
            ),
        )

        # Create routing
        await RoutingRow.create(root_ctx.db, endpoint_id, session_id)

        resp["sessionId"] = str(session_id)  # changed since API v5
        resp["sessionName"] = str(params["service_name"])
        resp["status"] = SessionStatus.PENDING.name
        resp["servicePorts"] = []
        resp["created"] = True

        # if not params["enqueue_only"]:
        #     app_ctx.pending_waits.add(current_task)
        #     max_wait = params["max_wait_seconds"]
        #     try:
        #         if max_wait > 0:
        #             with timeout(max_wait):
        #                 await start_event.wait()
        #         else:
        #             await start_event.wait()
        #     except asyncio.TimeoutError:
        #         resp["status"] = "TIMEOUT"
        #     else:
        #         await asyncio.sleep(0.5)
        #         async with root_ctx.db.begin_readonly_session() as db_sess:
        #             query = sa.select(KernelRow.status, KernelRow.service_ports).where(
        #                 (KernelRow.session_id == session_id)
        #                 & (KernelRow.cluster_role == DEFAULT_ROLE)
        #             )
        #             result = await db_sess.execute(query)
        #             row = result.first()
        #         if row.status == KernelStatus.RUNNING:
        #             resp["status"] = "RUNNING"
        #             for item in row.service_ports:
        #                 response_dict = {
        #                     "name": item["name"],
        #                     "protocol": item["protocol"],
        #                     "ports": item["container_ports"],
        #                 }
        #                 if "url_template" in item.keys():
        #                     response_dict["url_template"] = item["url_template"]
        #                 if "allowed_arguments" in item.keys():
        #                     response_dict["allowed_arguments"] = item["allowed_arguments"]
        #                 if "allowed_envs" in item.keys():
        #                     response_dict["allowed_envs"] = item["allowed_envs"]
        #                 resp["servicePorts"].append(response_dict)
        #         else:
        #             resp["status"] = row.status.name
    except asyncio.CancelledError:
        raise
    except BackendError:
        log.exception("GET_OR_CREATE: exception")
        raise
    except UnknownImageReference:
        raise UnknownImageReferenceError(f"Unknown image reference: {params['image']}")
    except Exception:
        log.exception("GET_OR_CREATE: unexpected error!")
        raise InternalServerError
    return web.json_response(resp, status=201)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["endpoint_id", "endpointId"]): tx.UUID | t.String,
        }
    ),
)
async def start(request: web.Request, params: Any) -> web.Response:
    access_key = request["keypair"]["access_key"]

    log.info("SERVE.START (email:{}, ak:{})", request["user"]["email"], access_key)
    return web.Response(status=204)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["endpoint_id", "endpointId"]): tx.UUID | t.String,
        }
    ),
)
async def stop(request: web.Request, params: Any) -> web.Response:
    access_key = request["keypair"]["access_key"]

    log.info("SERVE.STOP (email:{}, ak:{})", request["user"]["email"], access_key)
    return web.Response(status=204)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["endpoint_id", "endpointId"]): tx.UUID | t.String,
            tx.AliasedKey(["input_args", "inputArgs"], default=dict): t.Mapping(t.String, t.Any),
        }
    ),
)
async def invoke_serving(request: web.Request, params: Any) -> web.Response:
    access_key = request["keypair"]["access_key"]

    log.info("SERVE.INVOKE (email:{}, ak:{})", request["user"]["email"], access_key)
    return web.Response(status=204)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["service_id", "serviceId"]): tx.UUID,
        }
    ),
)
async def delete(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    # session_name = request.match_info["session_name"]
    service_id = params["service_id"]

    log.info("SERVE.DELETE (email:{}, ak:{})", request["user"]["email"], access_key)

    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    if requester_access_key != owner_access_key and request["user"]["role"] not in (
        UserRole.ADMIN,
        UserRole.SUPERADMIN,
    ):
        raise InsufficientPrivilege("You are not allowed to delete others's services")

    async with root_ctx.db.begin_session() as db_sess:
        # Delete endpoint first
        await db_sess.execute(sa.delete(EndpointRow).where(session=service_id))

        session = await SessionRow.get_session_with_kernels(service_id, db_session=db_sess)
    last_stat = await root_ctx.registry.destroy_session(
        session,
    )
    resp = {
        "stats": last_stat,
    }
    return web.json_response(resp, status=200)


@attrs.define(slots=True, auto_attribs=True, init=False)
class PrivateContext:
    database_ptask_group: aiotools.PersistentTaskGroup


async def init(app: web.Application) -> None:
    app_ctx: PrivateContext = app["services.context"]
    app_ctx.database_ptask_group = aiotools.PersistentTaskGroup()


async def shutdown(app: web.Application) -> None:
    app_ctx: PrivateContext = app["services.context"]
    await app_ctx.database_ptask_group.shutdown()


def create_app(
    default_cors_options: CORSOptions,
) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "services"
    app["api_versions"] = (4, 5)
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app["services.context"] = PrivateContext()
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    add_route = app.router.add_route
    root_resource = cors.add(app.router.add_resource(r""))
    cors.add(root_resource.add_route("GET", list_serve))
    cors.add(root_resource.add_route("POST", create))
    cors.add(root_resource.add_route("DELETE", delete))
    cors.add(add_route("GET", r"/_/info", get_info))
    cors.add(add_route("POST", r"/_/start", start))
    cors.add(add_route("POST", r"/_/stop", stop))
    cors.add(add_route("POST", r"/_/invoke", invoke_serving))
    return app, []



"""dynamic-resource-slots

Revision ID: f0f4ee907155
Revises: ff4bfca66bf8
Create Date: 2019-01-27 17:05:13.997279

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = "f0f4ee907155"
down_revision = "ff4bfca66bf8"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    connection = op.get_bind()
    op.alter_column(
        "kernels",
        "service_ports",
        existing_type=sa.JSON(),
        type_=postgresql.JSONB(),
        postgresql_using="CAST(service_ports AS jsonb)",
    )
    op.add_column(
        "agents",
        sa.Column(
            "available_slots",
            postgresql.JSONB(),
            nullable=False,
            server_default=sa.text("'{}'::jsonb"),
        ),
    )
    op.add_column(
        "agents",
        sa.Column(
            "occupied_slots",
            postgresql.JSONB(),
            nullable=False,
            server_default=sa.text("'{}'::jsonb"),
        ),
    )
    queries = """
    UPDATE agents SET available_slots = available_slots || json_build_object(
        'cuda.device', gpu_slots
    )::jsonb
    WHERE gpu_slots > 0;
    UPDATE agents SET available_slots = available_slots || json_build_object(
        'tpu.device', tpu_slots
    )::jsonb
    WHERE tpu_slots > 0;
    UPDATE agents SET occupied_slots = json_strip_nulls(json_build_object(
        'cpu', used_cpu_slots,
        'mem', used_mem_slots::text || 'g'
    ));
    UPDATE agents SET occupied_slots = occupied_slots || json_build_object(
        'cuda.device', used_gpu_slots
    )::jsonb
    WHERE used_gpu_slots > 0;
    UPDATE agents SET occupied_slots = occupied_slots || json_build_object(
        'tpu.device', used_tpu_slots
    )::jsonb
    WHERE used_tpu_slots > 0;
    """
    for query in queries.split(";"):
        if len(query.strip()) == 0:
            continue
        connection.execute(text(query))
    op.drop_column("agents", "cpu_slots")
    op.drop_column("agents", "mem_slots")
    op.drop_column("agents", "gpu_slots")
    op.drop_column("agents", "tpu_slots")
    op.drop_column("agents", "used_cpu_slots")
    op.drop_column("agents", "used_mem_slots")
    op.drop_column("agents", "used_gpu_slots")
    op.drop_column("agents", "used_tpu_slots")

    op.add_column(
        "kernels",
        sa.Column(
            "occupied_slots",
            postgresql.JSONB(),
            nullable=False,
            server_default=sa.text("'{}'::jsonb"),
        ),
    )
    op.add_column(
        "kernels",
        sa.Column(
            "occupied_shares",
            postgresql.JSONB(),
            nullable=False,
            server_default=sa.text("'{}'::jsonb"),
        ),
    )
    queries = """
    UPDATE kernels SET occupied_slots = json_build_object(
        'cpu', cpu_slot,
        'mem', mem_slot,
        'cuda.device', gpu_slot,
        'tpu.device', tpu_slot
    );
    UPDATE kernels SET occupied_shares = json_build_object(
        'cpu', cpu_set,
        'mem', mem_slot,
        'cuda.device', '{}'::json,
        'tpu.device', '{}'::json
    );
    """
    for query in queries.split(";"):
        if len(query.strip()) == 0:
            continue
        connection.execute(text(query))
    op.drop_column("kernels", "cpu_slot")
    op.drop_column("kernels", "mem_slot")
    op.drop_column("kernels", "gpu_slot")
    op.drop_column("kernels", "tpu_slot")
    op.drop_column("kernels", "cpu_set")
    op.drop_column("kernels", "gpu_set")
    op.drop_column("kernels", "tpu_set")
    # ### end Alembic commands ###


def downgrade():
    op.alter_column(
        "kernels",
        "service_ports",
        existing_type=postgresql.JSONB(),
        type_=sa.JSON(),
        postgresql_using="CAST(service_ports AS json)",
    )
    connection = op.get_bind()

    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "kernels",
        sa.Column("cpu_set", postgresql.ARRAY(sa.INTEGER()), autoincrement=False, nullable=True),
    )
    op.add_column(
        "kernels",
        sa.Column("gpu_set", postgresql.ARRAY(sa.INTEGER()), autoincrement=False, nullable=True),
    )
    op.add_column(
        "kernels",
        sa.Column("tpu_set", postgresql.ARRAY(sa.INTEGER()), autoincrement=False, nullable=True),
    )
    op.add_column(
        "kernels",
        sa.Column(
            "cpu_slot",
            postgresql.DOUBLE_PRECISION(precision=53),
            autoincrement=False,
            nullable=False,
            server_default="0",
        ),
    )
    op.add_column(
        "kernels",
        sa.Column("mem_slot", sa.BIGINT(), autoincrement=False, nullable=False, server_default="0"),
    )
    op.add_column(
        "kernels",
        sa.Column(
            "gpu_slot",
            postgresql.DOUBLE_PRECISION(precision=53),
            autoincrement=False,
            nullable=False,
            server_default="0",
        ),
    )
    op.add_column(
        "kernels",
        sa.Column(
            "tpu_slot",
            postgresql.DOUBLE_PRECISION(precision=53),
            autoincrement=False,
            nullable=False,
            server_default="0",
        ),
    )
    query = """
    UPDATE kernels
    SET
        cpu_set = (
            SELECT coalesce(array_agg(v::text::int), '{}')
            FROM json_array_elements((occupied_shares->>'cpu')::json) v),
        gpu_set = (
            SELECT coalesce(array_agg(v::text::int), '{}')
            FROM json_array_elements((occupied_shares->>'cuda.device')::json) v),
        tpu_set = (
            SELECT coalesce(array_agg(v::text::int), '{}')
            FROM json_array_elements((occupied_shares->>'tpu.device')::json) v)
    ;
    """
    connection.execute(text(query))
    query = """
    UPDATE kernels
    SET
        cpu_slot = coalesce((occupied_slots->>'cpu')::text::float, 0),
        mem_slot = coalesce((occupied_slots->>'mem')::text::bigint, 0),
        gpu_slot = coalesce((occupied_slots->>'cuda.device')::text::float, 0),
        tpu_slot = coalesce((occupied_slots->>'tpu.device')::text::float, 0)
    ;
    """
    connection.execute(text(query))
    op.drop_column("kernels", "occupied_shares")
    op.drop_column("kernels", "occupied_slots")

    op.add_column(
        "agents",
        sa.Column(
            "gpu_slots",
            postgresql.DOUBLE_PRECISION(precision=53),
            autoincrement=False,
            nullable=False,
            server_default="0",
        ),
    )
    op.add_column(
        "agents",
        sa.Column(
            "used_cpu_slots",
            postgresql.DOUBLE_PRECISION(precision=53),
            autoincrement=False,
            nullable=False,
            server_default="0",
        ),
    )
    op.add_column(
        "agents",
        sa.Column(
            "tpu_slots",
            postgresql.DOUBLE_PRECISION(precision=53),
            autoincrement=False,
            nullable=False,
            server_default="0",
        ),
    )
    op.add_column(
        "agents",
        sa.Column(
            "used_mem_slots", sa.BIGINT(), autoincrement=False, nullable=False, server_default="0"
        ),
    )
    op.add_column(
        "agents",
        sa.Column(
            "cpu_slots",
            postgresql.DOUBLE_PRECISION(precision=53),
            autoincrement=False,
            nullable=False,
            server_default="0",
        ),
    )
    op.add_column(
        "agents",
        sa.Column(
            "used_gpu_slots",
            postgresql.DOUBLE_PRECISION(precision=53),
            autoincrement=False,
            nullable=False,
            server_default="0",
        ),
    )
    op.add_column(
        "agents",
        sa.Column(
            "mem_slots", sa.BIGINT(), autoincrement=False, nullable=False, server_default="0"
        ),
    )
    op.add_column(
        "agents",
        sa.Column(
            "used_tpu_slots",
            postgresql.DOUBLE_PRECISION(precision=53),
            autoincrement=False,
            nullable=False,
            server_default="0",
        ),
    )
    query = """
    UPDATE agents
    SET
        cpu_slots = coalesce((available_slots->>'cpu')::text::float, 0),
        mem_slots = coalesce((available_slots->>'mem')::text::bigint, 0),
        gpu_slots = coalesce((available_slots->>'cuda.device')::text::float, 0),
        tpu_slots = coalesce((available_slots->>'tpu.device')::text::float, 0)
    ;
    """
    connection.execute(text(query))
    query = """
    UPDATE agents
    SET
        used_cpu_slots = coalesce((occupied_slots->>'cpu')::text::float, 0),
        used_mem_slots = coalesce((occupied_slots->>'mem')::text::bigint, 0),
        used_gpu_slots = coalesce((occupied_slots->>'cuda.device')::text::float, 0),
        used_tpu_slots = coalesce((occupied_slots->>'tpu.device')::text::float, 0)
    ;
    """
    connection.execute(text(query))
    op.drop_column("agents", "occupied_slots")
    op.drop_column("agents", "available_slots")
    # ### end Alembic commands ###



"""init

Revision ID: 5de06da3c2b5
Revises:
Create Date: 2017-06-08 15:08:23.166237

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import text

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "5de06da3c2b5"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.execute(text('CREATE EXTENSION IF NOT EXISTS "uuid-ossp";'))
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "keypairs",
        sa.Column("user_id", sa.Integer(), nullable=True),
        sa.Column("access_key", sa.String(length=20), nullable=False),
        sa.Column("secret_key", sa.String(length=40), nullable=True),
        sa.Column("is_active", sa.Boolean(), nullable=True),
        sa.Column("billing_plan", sa.String(), nullable=True),
        sa.Column(
            "created_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=True
        ),
        sa.Column("last_used", sa.DateTime(timezone=True), nullable=True),
        sa.Column("concurrency_limit", sa.Integer(), nullable=True),
        sa.Column("concurrency_used", sa.Integer(), nullable=True),
        sa.Column("rate_limit", sa.Integer(), nullable=True),
        sa.Column("num_queries", sa.Integer(), server_default="0", nullable=True),
        sa.PrimaryKeyConstraint("access_key"),
    )
    op.create_table(
        "kernels",
        sa.Column("sess_id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("lang", sa.String(length=64), nullable=True),
        sa.Column("access_key", sa.String(length=20), nullable=True),
        sa.Column(
            "created_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=True
        ),
        sa.Column("terminated_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("status", sa.String(), nullable=True),
        sa.Column("agent_id", sa.String(), nullable=True),
        sa.Column("container_id", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["access_key"],
            ["keypairs.access_key"],
        ),
        sa.PrimaryKeyConstraint("sess_id"),
    )
    op.create_table(
        "usage",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("access_key_id", sa.String(length=20), nullable=True),
        sa.Column("kernel_type", sa.String(), nullable=True),
        sa.Column("kernel_id", sa.String(), nullable=True),
        sa.Column("started_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("terminated_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("cpu_used", sa.Integer(), server_default="0", nullable=True),
        sa.Column("mem_used", sa.Integer(), server_default="0", nullable=True),
        sa.Column("io_used", sa.Integer(), server_default="0", nullable=True),
        sa.Column("net_used", sa.Integer(), server_default="0", nullable=True),
        sa.ForeignKeyConstraint(
            ["access_key_id"],
            ["keypairs.access_key"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("usage")
    op.drop_table("kernels")
    op.drop_table("keypairs")
    # ### end Alembic commands ###



"""add totp_activated_at column

Revision ID: 3efd66393bd0
Revises: ac4e179c57fe
Create Date: 2023-02-18 23:50:57.504752

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3efd66393bd0"
down_revision = "ac4e179c57fe"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "users", sa.Column("totp_activated_at", sa.DateTime(timezone=True), nullable=True)
    )


def downgrade():
    op.drop_column("users", "totp_activated_at")



"""add-status_data

Revision ID: 250e8656cf45
Revises: 57e717103287
Create Date: 2020-12-23 14:19:08.801283

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "250e8656cf45"
down_revision = "57e717103287"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "kernels", sa.Column("status_data", postgresql.JSONB(astext_type=sa.Text()), nullable=True)
    )


def downgrade():
    op.drop_column("kernels", "status_data")



"""Add local_rank column in Kernels table

Revision ID: 6129745f49b6
Revises: 360af8f33d4e
Create Date: 2022-10-22 15:58:01.518770

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "6129745f49b6"
down_revision = "360af8f33d4e"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "kernels",
        sa.Column("local_rank", sa.Integer, nullable=False, server_default=sa.text("0")),
    )


def downgrade():
    op.drop_column("kernels", "local_rank")



"""add a new session type INFERENCE

Revision ID: b5be363ab05c
Revises: cace152eefac
Create Date: 2023-02-15 15:24:29.112938

Reference: https://www.postgresql.org/message-id/CANu8FiwwBxZZGX23=Na_7bc4DZ-yzd_poKhaoPmN3+SHG08MAg@mail.gmail.com

"""
import textwrap

from alembic import op
from sqlalchemy.sql import text

from ai.backend.common.types import SessionTypes

# revision identifiers, used by Alembic.
revision = "b5be363ab05c"
down_revision = "cace152eefac"
branch_labels = None
depends_on = None

# select enum_range(null::sessiontypes);
typename = SessionTypes.__name__.lower()


def upgrade():
    op.execute(text("ALTER TYPE {} ADD VALUE IF NOT EXISTS 'INFERENCE';".format(typename)))
    op.execute(
        text(
            textwrap.dedent(
                """\
                ALTER TABLE sessions ALTER COLUMN session_type TYPE {} USING session_type::text::sessiontypes;
            """.format(
                    typename
                )
            )
        )
    )


def downgrade():
    conn = op.get_bind()
    conn.execute(
        text(
            "UPDATE sessions SET session_type = 'INTERACTIVE'::{x} WHERE session_type = 'INFERENCE'::{x};".format(
                x=typename
            )
        )
    )

    cursor = conn.execute(
        text(
            textwrap.dedent(
                """\
                SELECT t.typname, e.enumlabel, e.enumsortorder, e.enumtypid
                FROM pg_type t
                JOIN pg_enum e ON e.enumtypid = t.oid
                WHERE t.typtype = 'e'
                AND e.enumlabel = 'INFERENCE'
                ORDER BY 1, enumsortorder;
            """
            )
        )
    )

    if (row := next(cursor, None)) is None:
        return

    typname, enumlabel, enumsortorder, enumtypid = row

    conn.execute(
        text(
            textwrap.dedent(
                """\
                DELETE FROM pg_enum
                WHERE enumtypid = {enumtypid}
                AND enumlabel = 'INFERENCE';
            """.format(
                    enumtypid=enumtypid
                )
            )
        )
    )



"""add_resource_opts_in_kernels

Revision ID: 5b45f28d2cac
Revises: 9c89b9011872
Create Date: 2019-09-08 10:07:20.971662

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "5b45f28d2cac"
down_revision = "9c89b9011872"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "kernels",
        sa.Column("resource_opts", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("kernels", "resource_opts")
    # ### end Alembic commands ###
